#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.hw.dart import DART
from m1n1.hw.scaler import *
from m1n1.utils import *
import struct
import time
from PIL import Image, ImageDraw

SCALER_ADT = '/arm-io/scaler0'
DART_ADT = '/arm-io/dart-scaler0'

p.pmgr_adt_clocks_enable(DART_ADT)
p.pmgr_adt_clocks_enable(SCALER_ADT)

dart = DART.from_adt(u, DART_ADT)
dart.initialize()

scaler_base, _ = u.adt[SCALER_ADT].get_reg(0)
apiodma_base, _ = u.adt[SCALER_ADT].get_reg(1)
dpe_ctrl_base, _ = u.adt[SCALER_ADT].get_reg(2)

scaler = ScalerMainRegs(u, scaler_base)

def dpe_start():
    p.write32(dpe_ctrl_base + 0x400, 0x1)
    p.write32(dpe_ctrl_base + 0x404, 0x1)
    p.write32(dpe_ctrl_base + 0x438, 0xf)
    p.write32(dpe_ctrl_base + 0x43c, 0x5)
    p.write32(dpe_ctrl_base + 0x408, 0x1)
    p.write32(dpe_ctrl_base + 0x440, 0x5)
    p.write32(dpe_ctrl_base + 0x444, 0x4)
    p.write32(dpe_ctrl_base + 0x40c, 0x1)
    p.write32(dpe_ctrl_base + 0x448, 0x5)
    p.write32(dpe_ctrl_base + 0x44c, 0x5)
    p.write32(dpe_ctrl_base + 0x410, 0x1)
    p.write32(dpe_ctrl_base + 0x450, 0x7)
    p.write32(dpe_ctrl_base + 0x454, 0x7)
    p.write32(dpe_ctrl_base + 0x414, 0x1)
    p.write32(dpe_ctrl_base + 0x458, 0xd)
    p.write32(dpe_ctrl_base + 0x45c, 0xc)
    p.write32(dpe_ctrl_base + 0x418, 0x1)
    p.write32(dpe_ctrl_base + 0x460, 0x13)
    p.write32(dpe_ctrl_base + 0x464, 0x12)
    p.write32(dpe_ctrl_base + 0x41c, 0x1)
    p.write32(dpe_ctrl_base + 0x468, 0x9)
    p.write32(dpe_ctrl_base + 0x46c, 0xa)
    p.write32(dpe_ctrl_base + 0x420, 0x1)
    p.write32(dpe_ctrl_base + 0x470, 0x33)
    p.write32(dpe_ctrl_base + 0x474, 0x2c)
    p.write32(dpe_ctrl_base + 0x424, 0x1)
    p.write32(dpe_ctrl_base + 0x478, 0x15)
    p.write32(dpe_ctrl_base + 0x47c, 0x15)
    p.write32(dpe_ctrl_base + 0x428, 0x1)
    p.write32(dpe_ctrl_base + 0x480, 0xe)
    p.write32(dpe_ctrl_base + 0x484, 0x5)
    p.write32(dpe_ctrl_base + 0x42c, 0x1)
    p.write32(dpe_ctrl_base + 0x488, 0x27)
    p.write32(dpe_ctrl_base + 0x48c, 0x15)
    p.write32(dpe_ctrl_base + 0x430, 0x1)
    p.write32(dpe_ctrl_base + 0x490, 0x15)
    p.write32(dpe_ctrl_base + 0x494, 0xe)
    p.write32(dpe_ctrl_base + 0x434, 0x1)
    p.write32(dpe_ctrl_base + 0x498, 0x0)
    p.write32(dpe_ctrl_base + 0x49c, 0x0)
    p.write32(dpe_ctrl_base + 0x4, 0x1000)
    p.write32(dpe_ctrl_base + 0x0, 0x101)

def dpe_stop():
    p.write32(dpe_ctrl_base + 0x0, 0x103)
    while p.read32(dpe_ctrl_base + 0x0) & 0xC != 4:
        ...
    p.write32(dpe_ctrl_base + 0x0, p.read32(dpe_ctrl_base + 0x0) & 0xfffffffc)

print(f"Hardware version {scaler.HW_VERSION.val:08X}")

scaler.RESET = 1
scaler.RESET = 0

print(f"Hardware version after reset {scaler.HW_VERSION.val:08X}")

if len(sys.argv) < 3:
    print(f"Usage: {sys.argv[0]} input.png output.png")
    sys.exit(-1)

input_image_fn = sys.argv[1]
output_image_fn = sys.argv[2]

in_data = b''
with Image.open(input_image_fn) as im:
    in_W, in_H = im.size
    in_BYTESPP = 4
    in_STRIDE = in_W * in_BYTESPP
    in_SZ = in_W * in_H * in_BYTESPP

    for y in range(in_H):
        for x in range(in_W):
            r, g, b = im.getpixel((x, y))
            in_data += struct.pack("BBBB", r, g, b, 255)

out_W = in_W * 5
out_H = in_H * 3
out_BYTESPP = 4
out_STRIDE = out_W * out_BYTESPP
out_SZ = out_W * out_H * out_BYTESPP * 2  # HACK: double size for testing purposes

for i in range(in_W * in_H):
    in_data += struct.pack("<I", i & 0xFFFFFFFF)
# chexdump(in_data)

out_buf_phys = u.heap.memalign(0x4000, out_SZ)
iface.writemem(out_buf_phys, b'\xAA' * out_SZ)
out_buf_iova = dart.iomap(0, out_buf_phys, out_SZ)
print(f"Output buffer @ phys {out_buf_phys:016X} iova {out_buf_iova:016X}")

in_buf_phys = u.heap.memalign(0x4000, in_SZ)
iface.writemem(in_buf_phys, in_data)
in_buf_iova = dart.iomap(0, in_buf_phys, in_SZ)
print(f"Input buffer @ phys {in_buf_phys:016X} iova {in_buf_iova:016X}")
dart.dump_all()



dpe_start()

# reset CM
p.write32(scaler_base + 0x3800, 0x0)

# RDMA control
p.write32(scaler_base + 0x180, 0x1)
p.write32(scaler_base + 0x184, 0x1e)
p.write32(scaler_base + 0x188, 0x0)
p.write32(scaler_base + 0x18c, 0x0)
p.write32(scaler_base + 0x190, 0x0)

# transform config (flip/rotate)
scaler.FLIP_ROTATE.set()

# cache hints
scaler.CACHE_HINTS_THING0[0].val = 0x7d311
scaler.CACHE_HINTS_THING0[1].val = 0x7d311
scaler.CACHE_HINTS_THING0[2].val = 0x7d311
scaler.CACHE_HINTS_THING0[3].val = 0x7d311
scaler.CACHE_HINTS_THING2[0].val = 0xbd311
scaler.CACHE_HINTS_THING2[1].val = 0xbd311
scaler.CACHE_HINTS_THING2[2].val = 0xbd311
# scaler.CACHE_HINTS_THING2[3].val = 0xbd311
scaler.CACHE_HINTS_THING1[0].val = 0x707
scaler.CACHE_HINTS_THING1[1].val = 0x707
scaler.CACHE_HINTS_THING1[2].val = 0x707
scaler.CACHE_HINTS_THING1[3].val = 0x707
scaler.CACHE_HINTS_THING3[0].val = 0xc0bd307
scaler.CACHE_HINTS_THING3[1].val = 0xc0bd307
scaler.CACHE_HINTS_THING3[2].val = 0xc0bd307
# scaler.CACHE_HINTS_THING3[3].val = 0xc0bd307

# tunables
scaler.TUNABLES_THING0[0].val = 0x20
scaler.TUNABLES_THING0[1].val = 0x20
scaler.TUNABLES_THING0[2].val = 0x20
scaler.TUNABLES_THING0[3].val = 0x20
scaler.TUNABLES_THING1[0].val = 0x4000720
scaler.TUNABLES_THING1[1].val = 0x4000720
scaler.TUNABLES_THING1[2].val = 0x4000720
# scaler.TUNABLES_THING1[3].val = 0x4000720

# dest base addresses
scaler.DST_PLANE1_LO = 0
scaler.DST_PLANE1_HI = 0
scaler.DST_PLANE0_LO = out_buf_iova & 0xFFFFFFFF
scaler.DST_PLANE0_HI = out_buf_iova >> 32
scaler.DST_PLANE2_LO = 0
scaler.DST_PLANE2_HI = 0

# src base addresses
scaler.SRC_PLANE1_LO = 0
scaler.SRC_PLANE1_HI = 0
scaler.SRC_PLANE0_LO = in_buf_iova & 0xFFFFFFFF
scaler.SRC_PLANE0_HI = in_buf_iova >> 32
scaler.SRC_PLANE2_LO = 0
scaler.SRC_PLANE2_HI = 0

# dest stride
scaler.DST_PLANE1_STRIDE = 0
scaler.DST_PLANE0_STRIDE = out_STRIDE
scaler.DST_PLANE2_STRIDE = 0

# src stride
scaler.SRC_PLANE1_STRIDE = 0
scaler.SRC_PLANE0_STRIDE = in_STRIDE
scaler.SRC_PLANE2_STRIDE = 0

# dest offset
scaler.DST_PLANE1_OFFSET = 0
scaler.DST_PLANE0_OFFSET = 0
scaler.DST_PLANE2_OFFSET = 0

# src offset
scaler.SRC_PLANE1_OFFSET = 0
scaler.SRC_PLANE0_OFFSET = 0
scaler.SRC_PLANE2_OFFSET = 0

# dest sizes
scaler.DST_W = out_W
scaler.DST_H = out_H

scaler.DST_SIZE_THING3 = 0
scaler.DST_SIZE_THING6 = 0
scaler.DST_SIZE_THING2 = 0
scaler.DST_SIZE_THING5 = 0
scaler.DST_SIZE_THING4 = 0
scaler.DST_SIZE_THING7 = 0

# src sizes
scaler.SRC_W = in_W
scaler.SRC_H = in_H

scaler.SRC_SIZE_THING3 = 0
scaler.SRC_SIZE_THING6 = 0
scaler.SRC_SIZE_THING2 = 0
scaler.SRC_SIZE_THING5 = 0
scaler.SRC_SIZE_THING4 = 0
scaler.SRC_SIZE_THING7 = 0

# swizzling
scaler.SRC_SWIZZLE = 0x03020100
scaler.DST_SWIZZLE = 0x03020100

# WDMA control
p.write32(scaler_base + 0x280, 0x1)
p.write32(scaler_base + 0x284, 0x81e)
p.write32(scaler_base + 0x288, 0x800)
p.write32(scaler_base + 0x28c, 0x800)

# pixel averaging
scaler.PIXEL_AVERAGING = 0

# ASE enhancement
p.write32(scaler_base + 0x16800, 0x0)

# ASE 3x1 transform
p.write32(scaler_base + 0x16080, 0x0)
p.write32(scaler_base + 0x16084, 0xb710367)
p.write32(scaler_base + 0x16088, 0x128)

# ASE interpolation
p.write32(scaler_base + 0x16600, 0x15)

# ASE angle detect
p.write32(scaler_base + 0x16504, 0x2000500)
p.write32(scaler_base + 0x16508, 0x3200)
p.write32(scaler_base + 0x16534, 0x8)
p.write32(scaler_base + 0x1651c, 0x851400)
p.write32(scaler_base + 0x16568, 0x250500)
p.write32(scaler_base + 0x16588, 0x496513)

# ASE config
p.write32(scaler_base + 0x16000, 0x0)

# chroma upsampling
p.write32(scaler_base + 0x800, 0xc)

# chroma downsampling
p.write32(scaler_base + 0x900, 0x0)

# DDA init V???
scaler.SCALE_H_DDA_THING0 = 0
scaler.SCALE_H_DDA_THING2 = 0
scaler.SCALE_V_DDA_THING1 = 0

# vertical scaling
scaler.SCALE_V_RATIO_0 = int(in_H / out_H * 0x400000)
scaler.SCALE_V_RATIO_4 = 0  # XXX what does this do?
scaler.SCALE_V_RATIO_1 = 0  # XXX what does this do?
scaler.SCALE_V_RATIO_2 = 0  # XXX what does this set do?
scaler.SCALE_V_RATIO_3 = 0  # XXX what does this set do?
scaler.SCALE_V_RATIO_5 = 0  # XXX what does this set do?
scaler.SCALE_V_FLAGS.set(EN=1)

# XXX this is a random filter grabbed from a random trace
scaler.SCALE_FILTER_V_BLOCK0[0].val = 0x0
scaler.SCALE_FILTER_V_BLOCK0[1].val = 0x50005
scaler.SCALE_FILTER_V_BLOCK1[0].val = 0x50000
scaler.SCALE_FILTER_V_BLOCK0[2].val = 0xb000b
scaler.SCALE_FILTER_V_BLOCK0[3].val = 0x100010
scaler.SCALE_FILTER_V_BLOCK1[1].val = 0x10000b
scaler.SCALE_FILTER_V_BLOCK0[4].val = 0x140014
scaler.SCALE_FILTER_V_BLOCK0[5].val = 0x180018
scaler.SCALE_FILTER_V_BLOCK1[2].val = 0x180014
scaler.SCALE_FILTER_V_BLOCK0[6].val = 0x1c001c
scaler.SCALE_FILTER_V_BLOCK0[7].val = 0x200020
scaler.SCALE_FILTER_V_BLOCK1[3].val = 0x20001c
scaler.SCALE_FILTER_V_BLOCK0[8].val = 0x230023
scaler.SCALE_FILTER_V_BLOCK0[9].val = 0x260026
scaler.SCALE_FILTER_V_BLOCK1[4].val = 0x260023
scaler.SCALE_FILTER_V_BLOCK0[10].val = 0x290029
scaler.SCALE_FILTER_V_BLOCK0[11].val = 0x2c002c
scaler.SCALE_FILTER_V_BLOCK1[5].val = 0x2c0029
scaler.SCALE_FILTER_V_BLOCK0[12].val = 0x2e002e
scaler.SCALE_FILTER_V_BLOCK0[13].val = 0x300030
scaler.SCALE_FILTER_V_BLOCK1[6].val = 0x30002e
scaler.SCALE_FILTER_V_BLOCK0[14].val = 0x320032
scaler.SCALE_FILTER_V_BLOCK0[15].val = 0x330033
scaler.SCALE_FILTER_V_BLOCK1[7].val = 0x330032
scaler.SCALE_FILTER_V_BLOCK0[16].val = 0xff87ff87
scaler.SCALE_FILTER_V_BLOCK0[17].val = 0xff90ff90
scaler.SCALE_FILTER_V_BLOCK1[8].val = 0xff90ff87
scaler.SCALE_FILTER_V_BLOCK0[18].val = 0xff99ff99
scaler.SCALE_FILTER_V_BLOCK0[19].val = 0xffa1ffa1
scaler.SCALE_FILTER_V_BLOCK1[9].val = 0xffa1ff99
scaler.SCALE_FILTER_V_BLOCK0[20].val = 0xffaaffaa
scaler.SCALE_FILTER_V_BLOCK0[21].val = 0xffb2ffb2
scaler.SCALE_FILTER_V_BLOCK1[10].val = 0xffb2ffaa
scaler.SCALE_FILTER_V_BLOCK0[22].val = 0xffbaffba
scaler.SCALE_FILTER_V_BLOCK0[23].val = 0xffc2ffc2
scaler.SCALE_FILTER_V_BLOCK1[11].val = 0xffc2ffba
scaler.SCALE_FILTER_V_BLOCK0[24].val = 0xffcaffca
scaler.SCALE_FILTER_V_BLOCK0[25].val = 0xffd2ffd2
scaler.SCALE_FILTER_V_BLOCK1[12].val = 0xffd2ffca
scaler.SCALE_FILTER_V_BLOCK0[26].val = 0xffd9ffd9
scaler.SCALE_FILTER_V_BLOCK0[27].val = 0xffe0ffe0
scaler.SCALE_FILTER_V_BLOCK1[13].val = 0xffe0ffd9
scaler.SCALE_FILTER_V_BLOCK0[28].val = 0xffe7ffe7
scaler.SCALE_FILTER_V_BLOCK0[29].val = 0xffeeffee
scaler.SCALE_FILTER_V_BLOCK1[14].val = 0xffeeffe7
scaler.SCALE_FILTER_V_BLOCK0[30].val = 0xfff4fff4
scaler.SCALE_FILTER_V_BLOCK0[31].val = 0xfffafffa
scaler.SCALE_FILTER_V_BLOCK1[15].val = 0xfffafff4
scaler.SCALE_FILTER_V_BLOCK0[32].val = 0xff06ff06
scaler.SCALE_FILTER_V_BLOCK0[33].val = 0xff0cff0c
scaler.SCALE_FILTER_V_BLOCK1[16].val = 0xff0cff06
scaler.SCALE_FILTER_V_BLOCK0[34].val = 0xff13ff13
scaler.SCALE_FILTER_V_BLOCK0[35].val = 0xff1aff1a
scaler.SCALE_FILTER_V_BLOCK1[17].val = 0xff1aff13
scaler.SCALE_FILTER_V_BLOCK0[36].val = 0xff21ff21
scaler.SCALE_FILTER_V_BLOCK0[37].val = 0xff28ff28
scaler.SCALE_FILTER_V_BLOCK1[18].val = 0xff28ff21
scaler.SCALE_FILTER_V_BLOCK0[38].val = 0xff30ff30
scaler.SCALE_FILTER_V_BLOCK0[39].val = 0xff38ff38
scaler.SCALE_FILTER_V_BLOCK1[19].val = 0xff38ff30
scaler.SCALE_FILTER_V_BLOCK0[40].val = 0xff41ff41
scaler.SCALE_FILTER_V_BLOCK0[41].val = 0xff49ff49
scaler.SCALE_FILTER_V_BLOCK1[20].val = 0xff49ff41
scaler.SCALE_FILTER_V_BLOCK0[42].val = 0xff52ff52
scaler.SCALE_FILTER_V_BLOCK0[43].val = 0xff5bff5b
scaler.SCALE_FILTER_V_BLOCK1[21].val = 0xff5bff52
scaler.SCALE_FILTER_V_BLOCK0[44].val = 0xff63ff63
scaler.SCALE_FILTER_V_BLOCK0[45].val = 0xff6cff6c
scaler.SCALE_FILTER_V_BLOCK1[22].val = 0xff6cff63
scaler.SCALE_FILTER_V_BLOCK0[46].val = 0xff75ff75
scaler.SCALE_FILTER_V_BLOCK0[47].val = 0xff7eff7e
scaler.SCALE_FILTER_V_BLOCK1[23].val = 0xff7eff75
scaler.SCALE_FILTER_V_BLOCK0[48].val = 0xff02ff02
scaler.SCALE_FILTER_V_BLOCK0[49].val = 0xfefcfefc
scaler.SCALE_FILTER_V_BLOCK1[24].val = 0xfefcff02
scaler.SCALE_FILTER_V_BLOCK0[50].val = 0xfef7fef7
scaler.SCALE_FILTER_V_BLOCK0[51].val = 0xfef3fef3
scaler.SCALE_FILTER_V_BLOCK1[25].val = 0xfef3fef7
scaler.SCALE_FILTER_V_BLOCK0[52].val = 0xfeeffeef
scaler.SCALE_FILTER_V_BLOCK0[53].val = 0xfeedfeed
scaler.SCALE_FILTER_V_BLOCK1[26].val = 0xfeedfeef
scaler.SCALE_FILTER_V_BLOCK0[54].val = 0xfeecfeec
scaler.SCALE_FILTER_V_BLOCK0[55].val = 0xfeebfeeb
scaler.SCALE_FILTER_V_BLOCK1[27].val = 0xfeebfeec
scaler.SCALE_FILTER_V_BLOCK0[56].val = 0xfeebfeeb
scaler.SCALE_FILTER_V_BLOCK0[57].val = 0xfeecfeec
scaler.SCALE_FILTER_V_BLOCK1[28].val = 0xfeecfeeb
scaler.SCALE_FILTER_V_BLOCK0[58].val = 0xfeeefeee
scaler.SCALE_FILTER_V_BLOCK0[59].val = 0xfef1fef1
scaler.SCALE_FILTER_V_BLOCK1[29].val = 0xfef1feee
scaler.SCALE_FILTER_V_BLOCK0[60].val = 0xfef4fef4
scaler.SCALE_FILTER_V_BLOCK0[61].val = 0xfef8fef8
scaler.SCALE_FILTER_V_BLOCK1[30].val = 0xfef8fef4
scaler.SCALE_FILTER_V_BLOCK0[62].val = 0xfefcfefc
scaler.SCALE_FILTER_V_BLOCK0[63].val = 0xff01ff01
scaler.SCALE_FILTER_V_BLOCK1[31].val = 0xff01fefc
scaler.SCALE_FILTER_V_BLOCK0[64].val = 0x0
scaler.SCALE_FILTER_V_BLOCK0[65].val = 0xffe7ffe7
scaler.SCALE_FILTER_V_BLOCK1[32].val = 0xffe70000
scaler.SCALE_FILTER_V_BLOCK0[66].val = 0xffcfffcf
scaler.SCALE_FILTER_V_BLOCK0[67].val = 0xffb9ffb9
scaler.SCALE_FILTER_V_BLOCK1[33].val = 0xffb9ffcf
scaler.SCALE_FILTER_V_BLOCK0[68].val = 0xffa4ffa4
scaler.SCALE_FILTER_V_BLOCK0[69].val = 0xff90ff90
scaler.SCALE_FILTER_V_BLOCK1[34].val = 0xff90ffa4
scaler.SCALE_FILTER_V_BLOCK0[70].val = 0xff7dff7d
scaler.SCALE_FILTER_V_BLOCK0[71].val = 0xff6bff6b
scaler.SCALE_FILTER_V_BLOCK1[35].val = 0xff6bff7d
scaler.SCALE_FILTER_V_BLOCK0[72].val = 0xff5bff5b
scaler.SCALE_FILTER_V_BLOCK0[73].val = 0xff4cff4c
scaler.SCALE_FILTER_V_BLOCK1[36].val = 0xff4cff5b
scaler.SCALE_FILTER_V_BLOCK0[74].val = 0xff3eff3e
scaler.SCALE_FILTER_V_BLOCK0[75].val = 0xff31ff31
scaler.SCALE_FILTER_V_BLOCK1[37].val = 0xff31ff3e
scaler.SCALE_FILTER_V_BLOCK0[76].val = 0xff26ff26
scaler.SCALE_FILTER_V_BLOCK0[77].val = 0xff1bff1b
scaler.SCALE_FILTER_V_BLOCK1[38].val = 0xff1bff26
scaler.SCALE_FILTER_V_BLOCK0[78].val = 0xff12ff12
scaler.SCALE_FILTER_V_BLOCK0[79].val = 0xff0aff0a
scaler.SCALE_FILTER_V_BLOCK1[39].val = 0xff0aff12
scaler.SCALE_FILTER_V_BLOCK0[80].val = 0x2210221
scaler.SCALE_FILTER_V_BLOCK0[81].val = 0x1f901f9
scaler.SCALE_FILTER_V_BLOCK1[40].val = 0x1f90221
scaler.SCALE_FILTER_V_BLOCK0[82].val = 0x1d001d0
scaler.SCALE_FILTER_V_BLOCK0[83].val = 0x1a901a9
scaler.SCALE_FILTER_V_BLOCK1[41].val = 0x1a901d0
scaler.SCALE_FILTER_V_BLOCK0[84].val = 0x1820182
scaler.SCALE_FILTER_V_BLOCK0[85].val = 0x15d015d
scaler.SCALE_FILTER_V_BLOCK1[42].val = 0x15d0182
scaler.SCALE_FILTER_V_BLOCK0[86].val = 0x1380138
scaler.SCALE_FILTER_V_BLOCK0[87].val = 0x1140114
scaler.SCALE_FILTER_V_BLOCK1[43].val = 0x1140138
scaler.SCALE_FILTER_V_BLOCK0[88].val = 0xf100f1
scaler.SCALE_FILTER_V_BLOCK0[89].val = 0xcf00cf
scaler.SCALE_FILTER_V_BLOCK1[44].val = 0xcf00f1
scaler.SCALE_FILTER_V_BLOCK0[90].val = 0xae00ae
scaler.SCALE_FILTER_V_BLOCK0[91].val = 0x8e008e
scaler.SCALE_FILTER_V_BLOCK1[45].val = 0x8e00ae
scaler.SCALE_FILTER_V_BLOCK0[92].val = 0x6f006f
scaler.SCALE_FILTER_V_BLOCK0[93].val = 0x520052
scaler.SCALE_FILTER_V_BLOCK1[46].val = 0x52006f
scaler.SCALE_FILTER_V_BLOCK0[94].val = 0x350035
scaler.SCALE_FILTER_V_BLOCK0[95].val = 0x1a001a
scaler.SCALE_FILTER_V_BLOCK1[47].val = 0x1a0035
scaler.SCALE_FILTER_V_BLOCK0[96].val = 0x4e404e4
scaler.SCALE_FILTER_V_BLOCK0[97].val = 0x4b804b8
scaler.SCALE_FILTER_V_BLOCK1[48].val = 0x4b804e4
scaler.SCALE_FILTER_V_BLOCK0[98].val = 0x48b048b
scaler.SCALE_FILTER_V_BLOCK0[99].val = 0x45f045f
scaler.SCALE_FILTER_V_BLOCK1[49].val = 0x45f048b
scaler.SCALE_FILTER_V_BLOCK0[100].val = 0x4320432
scaler.SCALE_FILTER_V_BLOCK0[101].val = 0x4050405
scaler.SCALE_FILTER_V_BLOCK1[50].val = 0x4050432
scaler.SCALE_FILTER_V_BLOCK0[102].val = 0x3d803d8
scaler.SCALE_FILTER_V_BLOCK0[103].val = 0x3ab03ab
scaler.SCALE_FILTER_V_BLOCK1[51].val = 0x3ab03d8
scaler.SCALE_FILTER_V_BLOCK0[104].val = 0x37e037e
scaler.SCALE_FILTER_V_BLOCK0[105].val = 0x3510351
scaler.SCALE_FILTER_V_BLOCK1[52].val = 0x351037e
scaler.SCALE_FILTER_V_BLOCK0[106].val = 0x3240324
scaler.SCALE_FILTER_V_BLOCK0[107].val = 0x2f802f8
scaler.SCALE_FILTER_V_BLOCK1[53].val = 0x2f80324
scaler.SCALE_FILTER_V_BLOCK0[108].val = 0x2cc02cc
scaler.SCALE_FILTER_V_BLOCK0[109].val = 0x2a102a1
scaler.SCALE_FILTER_V_BLOCK1[54].val = 0x2a102cc
scaler.SCALE_FILTER_V_BLOCK0[110].val = 0x2760276
scaler.SCALE_FILTER_V_BLOCK0[111].val = 0x24b024b
scaler.SCALE_FILTER_V_BLOCK1[55].val = 0x24b0276
scaler.SCALE_FILTER_V_BLOCK0[112].val = 0x73b073b
scaler.SCALE_FILTER_V_BLOCK0[113].val = 0x71e071e
scaler.SCALE_FILTER_V_BLOCK1[56].val = 0x71e073b
scaler.SCALE_FILTER_V_BLOCK0[114].val = 0x7000700
scaler.SCALE_FILTER_V_BLOCK0[115].val = 0x6e106e1
scaler.SCALE_FILTER_V_BLOCK1[57].val = 0x6e10700
scaler.SCALE_FILTER_V_BLOCK0[116].val = 0x6c006c0
scaler.SCALE_FILTER_V_BLOCK0[117].val = 0x69e069e
scaler.SCALE_FILTER_V_BLOCK1[58].val = 0x69e06c0
scaler.SCALE_FILTER_V_BLOCK0[118].val = 0x67a067a
scaler.SCALE_FILTER_V_BLOCK0[119].val = 0x6560656
scaler.SCALE_FILTER_V_BLOCK1[59].val = 0x656067a
scaler.SCALE_FILTER_V_BLOCK0[120].val = 0x6300630
scaler.SCALE_FILTER_V_BLOCK0[121].val = 0x6090609
scaler.SCALE_FILTER_V_BLOCK1[60].val = 0x6090630
scaler.SCALE_FILTER_V_BLOCK0[122].val = 0x5e205e2
scaler.SCALE_FILTER_V_BLOCK0[123].val = 0x5b905b9
scaler.SCALE_FILTER_V_BLOCK1[61].val = 0x5b905e2
scaler.SCALE_FILTER_V_BLOCK0[124].val = 0x5900590
scaler.SCALE_FILTER_V_BLOCK0[125].val = 0x5660566
scaler.SCALE_FILTER_V_BLOCK1[62].val = 0x5660590
scaler.SCALE_FILTER_V_BLOCK0[126].val = 0x53b053b
scaler.SCALE_FILTER_V_BLOCK0[127].val = 0x5100510
scaler.SCALE_FILTER_V_BLOCK1[63].val = 0x510053b
scaler.SCALE_FILTER_V_BLOCK0[128].val = 0x82c082c
scaler.SCALE_FILTER_V_BLOCK0[129].val = 0x82b082b
scaler.SCALE_FILTER_V_BLOCK1[64].val = 0x82b082c
scaler.SCALE_FILTER_V_BLOCK0[130].val = 0x8280828
scaler.SCALE_FILTER_V_BLOCK0[131].val = 0x8200820
scaler.SCALE_FILTER_V_BLOCK1[65].val = 0x8200828
scaler.SCALE_FILTER_V_BLOCK0[132].val = 0x81b081b
scaler.SCALE_FILTER_V_BLOCK0[133].val = 0x8130813
scaler.SCALE_FILTER_V_BLOCK1[66].val = 0x813081b
scaler.SCALE_FILTER_V_BLOCK0[134].val = 0x8080808
scaler.SCALE_FILTER_V_BLOCK0[135].val = 0x7fc07fc
scaler.SCALE_FILTER_V_BLOCK1[67].val = 0x7fc0808
scaler.SCALE_FILTER_V_BLOCK0[136].val = 0x7ed07ed
scaler.SCALE_FILTER_V_BLOCK0[137].val = 0x7dd07dd
scaler.SCALE_FILTER_V_BLOCK1[68].val = 0x7dd07ed
scaler.SCALE_FILTER_V_BLOCK0[138].val = 0x7cb07cb
scaler.SCALE_FILTER_V_BLOCK0[139].val = 0x7b607b6
scaler.SCALE_FILTER_V_BLOCK1[69].val = 0x7b607cb
scaler.SCALE_FILTER_V_BLOCK0[140].val = 0x7a207a2
scaler.SCALE_FILTER_V_BLOCK0[141].val = 0x78a078a
scaler.SCALE_FILTER_V_BLOCK1[70].val = 0x78a07a2
scaler.SCALE_FILTER_V_BLOCK0[142].val = 0x7710771
scaler.SCALE_FILTER_V_BLOCK0[143].val = 0x7570757
scaler.SCALE_FILTER_V_BLOCK1[71].val = 0x7570771
scaler.SCALE_FILTER_V_BLOCK0[144].val = 0x73d073d
scaler.SCALE_FILTER_V_BLOCK0[145].val = 0x7570757
scaler.SCALE_FILTER_V_BLOCK1[72].val = 0x757073d
scaler.SCALE_FILTER_V_BLOCK0[146].val = 0x7710771
scaler.SCALE_FILTER_V_BLOCK0[147].val = 0x78a078a
scaler.SCALE_FILTER_V_BLOCK1[73].val = 0x78a0771
scaler.SCALE_FILTER_V_BLOCK0[148].val = 0x7a207a2
scaler.SCALE_FILTER_V_BLOCK0[149].val = 0x7b607b6
scaler.SCALE_FILTER_V_BLOCK1[74].val = 0x7b607a2
scaler.SCALE_FILTER_V_BLOCK0[150].val = 0x7cb07cb
scaler.SCALE_FILTER_V_BLOCK0[151].val = 0x7dd07dd
scaler.SCALE_FILTER_V_BLOCK1[75].val = 0x7dd07cb
scaler.SCALE_FILTER_V_BLOCK0[152].val = 0x7ed07ed
scaler.SCALE_FILTER_V_BLOCK0[153].val = 0x7fc07fc
scaler.SCALE_FILTER_V_BLOCK1[76].val = 0x7fc07ed
scaler.SCALE_FILTER_V_BLOCK0[154].val = 0x8080808
scaler.SCALE_FILTER_V_BLOCK0[155].val = 0x8130813
scaler.SCALE_FILTER_V_BLOCK1[77].val = 0x8130808
scaler.SCALE_FILTER_V_BLOCK0[156].val = 0x81b081b
scaler.SCALE_FILTER_V_BLOCK0[157].val = 0x8200820
scaler.SCALE_FILTER_V_BLOCK1[78].val = 0x820081b
scaler.SCALE_FILTER_V_BLOCK0[158].val = 0x8280828
scaler.SCALE_FILTER_V_BLOCK0[159].val = 0x82b082b
scaler.SCALE_FILTER_V_BLOCK1[79].val = 0x82b0828
scaler.SCALE_FILTER_V_BLOCK0[160].val = 0x4e404e4
scaler.SCALE_FILTER_V_BLOCK0[161].val = 0x5100510
scaler.SCALE_FILTER_V_BLOCK1[80].val = 0x51004e4
scaler.SCALE_FILTER_V_BLOCK0[162].val = 0x53b053b
scaler.SCALE_FILTER_V_BLOCK0[163].val = 0x5660566
scaler.SCALE_FILTER_V_BLOCK1[81].val = 0x566053b
scaler.SCALE_FILTER_V_BLOCK0[164].val = 0x5900590
scaler.SCALE_FILTER_V_BLOCK0[165].val = 0x5b905b9
scaler.SCALE_FILTER_V_BLOCK1[82].val = 0x5b90590
scaler.SCALE_FILTER_V_BLOCK0[166].val = 0x5e205e2
scaler.SCALE_FILTER_V_BLOCK0[167].val = 0x6090609
scaler.SCALE_FILTER_V_BLOCK1[83].val = 0x60905e2
scaler.SCALE_FILTER_V_BLOCK0[168].val = 0x6300630
scaler.SCALE_FILTER_V_BLOCK0[169].val = 0x6560656
scaler.SCALE_FILTER_V_BLOCK1[84].val = 0x6560630
scaler.SCALE_FILTER_V_BLOCK0[170].val = 0x67a067a
scaler.SCALE_FILTER_V_BLOCK0[171].val = 0x69e069e
scaler.SCALE_FILTER_V_BLOCK1[85].val = 0x69e067a
scaler.SCALE_FILTER_V_BLOCK0[172].val = 0x6c006c0
scaler.SCALE_FILTER_V_BLOCK0[173].val = 0x6e106e1
scaler.SCALE_FILTER_V_BLOCK1[86].val = 0x6e106c0
scaler.SCALE_FILTER_V_BLOCK0[174].val = 0x7000700
scaler.SCALE_FILTER_V_BLOCK0[175].val = 0x71e071e
scaler.SCALE_FILTER_V_BLOCK1[87].val = 0x71e0700
scaler.SCALE_FILTER_V_BLOCK0[176].val = 0x2210221
scaler.SCALE_FILTER_V_BLOCK0[177].val = 0x24b024b
scaler.SCALE_FILTER_V_BLOCK1[88].val = 0x24b0221
scaler.SCALE_FILTER_V_BLOCK0[178].val = 0x2760276
scaler.SCALE_FILTER_V_BLOCK0[179].val = 0x2a102a1
scaler.SCALE_FILTER_V_BLOCK1[89].val = 0x2a10276
scaler.SCALE_FILTER_V_BLOCK0[180].val = 0x2cc02cc
scaler.SCALE_FILTER_V_BLOCK0[181].val = 0x2f802f8
scaler.SCALE_FILTER_V_BLOCK1[90].val = 0x2f802cc
scaler.SCALE_FILTER_V_BLOCK0[182].val = 0x3240324
scaler.SCALE_FILTER_V_BLOCK0[183].val = 0x3510351
scaler.SCALE_FILTER_V_BLOCK1[91].val = 0x3510324
scaler.SCALE_FILTER_V_BLOCK0[184].val = 0x37e037e
scaler.SCALE_FILTER_V_BLOCK0[185].val = 0x3ab03ab
scaler.SCALE_FILTER_V_BLOCK1[92].val = 0x3ab037e
scaler.SCALE_FILTER_V_BLOCK0[186].val = 0x3d803d8
scaler.SCALE_FILTER_V_BLOCK0[187].val = 0x4050405
scaler.SCALE_FILTER_V_BLOCK1[93].val = 0x40503d8
scaler.SCALE_FILTER_V_BLOCK0[188].val = 0x4320432
scaler.SCALE_FILTER_V_BLOCK0[189].val = 0x45f045f
scaler.SCALE_FILTER_V_BLOCK1[94].val = 0x45f0432
scaler.SCALE_FILTER_V_BLOCK0[190].val = 0x48b048b
scaler.SCALE_FILTER_V_BLOCK0[191].val = 0x4b804b8
scaler.SCALE_FILTER_V_BLOCK1[95].val = 0x4b8048b
scaler.SCALE_FILTER_V_BLOCK0[192].val = 0x0
scaler.SCALE_FILTER_V_BLOCK0[193].val = 0x1a001a
scaler.SCALE_FILTER_V_BLOCK1[96].val = 0x1a0000
scaler.SCALE_FILTER_V_BLOCK0[194].val = 0x350035
scaler.SCALE_FILTER_V_BLOCK0[195].val = 0x520052
scaler.SCALE_FILTER_V_BLOCK1[97].val = 0x520035
scaler.SCALE_FILTER_V_BLOCK0[196].val = 0x6f006f
scaler.SCALE_FILTER_V_BLOCK0[197].val = 0x8e008e
scaler.SCALE_FILTER_V_BLOCK1[98].val = 0x8e006f
scaler.SCALE_FILTER_V_BLOCK0[198].val = 0xae00ae
scaler.SCALE_FILTER_V_BLOCK0[199].val = 0xcf00cf
scaler.SCALE_FILTER_V_BLOCK1[99].val = 0xcf00ae
scaler.SCALE_FILTER_V_BLOCK0[200].val = 0xf100f1
scaler.SCALE_FILTER_V_BLOCK0[201].val = 0x1140114
scaler.SCALE_FILTER_V_BLOCK1[100].val = 0x11400f1
scaler.SCALE_FILTER_V_BLOCK0[202].val = 0x1380138
scaler.SCALE_FILTER_V_BLOCK0[203].val = 0x15d015d
scaler.SCALE_FILTER_V_BLOCK1[101].val = 0x15d0138
scaler.SCALE_FILTER_V_BLOCK0[204].val = 0x1820182
scaler.SCALE_FILTER_V_BLOCK0[205].val = 0x1a901a9
scaler.SCALE_FILTER_V_BLOCK1[102].val = 0x1a90182
scaler.SCALE_FILTER_V_BLOCK0[206].val = 0x1d001d0
scaler.SCALE_FILTER_V_BLOCK0[207].val = 0x1f901f9
scaler.SCALE_FILTER_V_BLOCK1[103].val = 0x1f901d0
scaler.SCALE_FILTER_V_BLOCK0[208].val = 0xff02ff02
scaler.SCALE_FILTER_V_BLOCK0[209].val = 0xff0aff0a
scaler.SCALE_FILTER_V_BLOCK1[104].val = 0xff0aff02
scaler.SCALE_FILTER_V_BLOCK0[210].val = 0xff12ff12
scaler.SCALE_FILTER_V_BLOCK0[211].val = 0xff1bff1b
scaler.SCALE_FILTER_V_BLOCK1[105].val = 0xff1bff12
scaler.SCALE_FILTER_V_BLOCK0[212].val = 0xff26ff26
scaler.SCALE_FILTER_V_BLOCK0[213].val = 0xff31ff31
scaler.SCALE_FILTER_V_BLOCK1[106].val = 0xff31ff26
scaler.SCALE_FILTER_V_BLOCK0[214].val = 0xff3eff3e
scaler.SCALE_FILTER_V_BLOCK0[215].val = 0xff4cff4c
scaler.SCALE_FILTER_V_BLOCK1[107].val = 0xff4cff3e
scaler.SCALE_FILTER_V_BLOCK0[216].val = 0xff5bff5b
scaler.SCALE_FILTER_V_BLOCK0[218].val = 0xff6bff6b
scaler.SCALE_FILTER_V_BLOCK1[108].val = 0xff6bff5b
scaler.SCALE_FILTER_V_BLOCK0[218].val = 0xff7dff7d
scaler.SCALE_FILTER_V_BLOCK0[219].val = 0xff90ff90
scaler.SCALE_FILTER_V_BLOCK1[109].val = 0xff90ff7d
scaler.SCALE_FILTER_V_BLOCK0[220].val = 0xffa4ffa4
scaler.SCALE_FILTER_V_BLOCK0[221].val = 0xffb9ffb9
scaler.SCALE_FILTER_V_BLOCK1[110].val = 0xffb9ffa4
scaler.SCALE_FILTER_V_BLOCK0[222].val = 0xffcfffcf
scaler.SCALE_FILTER_V_BLOCK0[223].val = 0xffe7ffe7
scaler.SCALE_FILTER_V_BLOCK1[111].val = 0xffe7ffcf
scaler.SCALE_FILTER_V_BLOCK0[224].val = 0xff06ff06
scaler.SCALE_FILTER_V_BLOCK0[225].val = 0xff01ff01
scaler.SCALE_FILTER_V_BLOCK1[112].val = 0xff01ff06
scaler.SCALE_FILTER_V_BLOCK0[226].val = 0xfefcfefc
scaler.SCALE_FILTER_V_BLOCK0[227].val = 0xfef8fef8
scaler.SCALE_FILTER_V_BLOCK1[113].val = 0xfef8fefc
scaler.SCALE_FILTER_V_BLOCK0[228].val = 0xfef4fef4
scaler.SCALE_FILTER_V_BLOCK0[229].val = 0xfef1fef1
scaler.SCALE_FILTER_V_BLOCK1[114].val = 0xfef1fef4
scaler.SCALE_FILTER_V_BLOCK0[230].val = 0xfeeefeee
scaler.SCALE_FILTER_V_BLOCK0[231].val = 0xfeecfeec
scaler.SCALE_FILTER_V_BLOCK1[115].val = 0xfeecfeee
scaler.SCALE_FILTER_V_BLOCK0[232].val = 0xfeebfeeb
scaler.SCALE_FILTER_V_BLOCK0[233].val = 0xfeebfeeb
scaler.SCALE_FILTER_V_BLOCK1[116].val = 0xfeebfeeb
scaler.SCALE_FILTER_V_BLOCK0[234].val = 0xfeecfeec
scaler.SCALE_FILTER_V_BLOCK0[235].val = 0xfeedfeed
scaler.SCALE_FILTER_V_BLOCK1[117].val = 0xfeedfeec
scaler.SCALE_FILTER_V_BLOCK0[236].val = 0xfeeffeef
scaler.SCALE_FILTER_V_BLOCK0[237].val = 0xfef3fef3
scaler.SCALE_FILTER_V_BLOCK1[118].val = 0xfef3feef
scaler.SCALE_FILTER_V_BLOCK0[238].val = 0xfef7fef7
scaler.SCALE_FILTER_V_BLOCK0[239].val = 0xfefcfefc
scaler.SCALE_FILTER_V_BLOCK1[119].val = 0xfefcfef7
scaler.SCALE_FILTER_V_BLOCK0[240].val = 0xff87ff87
scaler.SCALE_FILTER_V_BLOCK0[241].val = 0xff7eff7e
scaler.SCALE_FILTER_V_BLOCK1[120].val = 0xff7eff87
scaler.SCALE_FILTER_V_BLOCK0[242].val = 0xff75ff75
scaler.SCALE_FILTER_V_BLOCK0[243].val = 0xff6cff6c
scaler.SCALE_FILTER_V_BLOCK1[121].val = 0xff6cff75
scaler.SCALE_FILTER_V_BLOCK0[244].val = 0xff63ff63
scaler.SCALE_FILTER_V_BLOCK0[245].val = 0xff5bff5b
scaler.SCALE_FILTER_V_BLOCK1[122].val = 0xff5bff63
scaler.SCALE_FILTER_V_BLOCK0[246].val = 0xff52ff52
scaler.SCALE_FILTER_V_BLOCK0[247].val = 0xff49ff49
scaler.SCALE_FILTER_V_BLOCK1[123].val = 0xff49ff52
scaler.SCALE_FILTER_V_BLOCK0[248].val = 0xff41ff41
scaler.SCALE_FILTER_V_BLOCK0[249].val = 0xff38ff38
scaler.SCALE_FILTER_V_BLOCK1[124].val = 0xff38ff41
scaler.SCALE_FILTER_V_BLOCK0[250].val = 0xff30ff30
scaler.SCALE_FILTER_V_BLOCK0[251].val = 0xff28ff28
scaler.SCALE_FILTER_V_BLOCK1[125].val = 0xff28ff30
scaler.SCALE_FILTER_V_BLOCK0[252].val = 0xff21ff21
scaler.SCALE_FILTER_V_BLOCK0[253].val = 0xff1aff1a
scaler.SCALE_FILTER_V_BLOCK1[126].val = 0xff1aff21
scaler.SCALE_FILTER_V_BLOCK0[254].val = 0xff13ff13
scaler.SCALE_FILTER_V_BLOCK0[255].val = 0xff0cff0c
scaler.SCALE_FILTER_V_BLOCK1[127].val = 0xff0cff13
scaler.SCALE_FILTER_V_BLOCK0[256].val = 0x0
scaler.SCALE_FILTER_V_BLOCK0[257].val = 0xfffafffa
scaler.SCALE_FILTER_V_BLOCK1[128].val = 0xfffa0000
scaler.SCALE_FILTER_V_BLOCK0[258].val = 0xfff4fff4
scaler.SCALE_FILTER_V_BLOCK0[259].val = 0xffeeffee
scaler.SCALE_FILTER_V_BLOCK1[129].val = 0xffeefff4
scaler.SCALE_FILTER_V_BLOCK0[260].val = 0xffe7ffe7
scaler.SCALE_FILTER_V_BLOCK0[261].val = 0xffe0ffe0
scaler.SCALE_FILTER_V_BLOCK1[130].val = 0xffe0ffe7
scaler.SCALE_FILTER_V_BLOCK0[262].val = 0xffd9ffd9
scaler.SCALE_FILTER_V_BLOCK0[263].val = 0xffd2ffd2
scaler.SCALE_FILTER_V_BLOCK1[131].val = 0xffd2ffd9
scaler.SCALE_FILTER_V_BLOCK0[264].val = 0xffcaffca
scaler.SCALE_FILTER_V_BLOCK0[265].val = 0xffc2ffc2
scaler.SCALE_FILTER_V_BLOCK1[132].val = 0xffc2ffca
scaler.SCALE_FILTER_V_BLOCK0[266].val = 0xffbaffba
scaler.SCALE_FILTER_V_BLOCK0[267].val = 0xffb2ffb2
scaler.SCALE_FILTER_V_BLOCK1[133].val = 0xffb2ffba
scaler.SCALE_FILTER_V_BLOCK0[268].val = 0xffaaffaa
scaler.SCALE_FILTER_V_BLOCK0[269].val = 0xffa1ffa1
scaler.SCALE_FILTER_V_BLOCK1[134].val = 0xffa1ffaa
scaler.SCALE_FILTER_V_BLOCK0[270].val = 0xff99ff99
scaler.SCALE_FILTER_V_BLOCK0[271].val = 0xff90ff90
scaler.SCALE_FILTER_V_BLOCK1[135].val = 0xff90ff99
scaler.SCALE_FILTER_V_BLOCK0[272].val = 0x340034
scaler.SCALE_FILTER_V_BLOCK0[273].val = 0x330033
scaler.SCALE_FILTER_V_BLOCK1[136].val = 0x330034
scaler.SCALE_FILTER_V_BLOCK0[274].val = 0x320032
scaler.SCALE_FILTER_V_BLOCK0[275].val = 0x300030
scaler.SCALE_FILTER_V_BLOCK1[137].val = 0x300032
scaler.SCALE_FILTER_V_BLOCK0[276].val = 0x2e002e
scaler.SCALE_FILTER_V_BLOCK0[277].val = 0x2c002c
scaler.SCALE_FILTER_V_BLOCK1[138].val = 0x2c002e
scaler.SCALE_FILTER_V_BLOCK0[278].val = 0x290029
scaler.SCALE_FILTER_V_BLOCK0[279].val = 0x260026
scaler.SCALE_FILTER_V_BLOCK1[139].val = 0x260029
scaler.SCALE_FILTER_V_BLOCK0[280].val = 0x230023
scaler.SCALE_FILTER_V_BLOCK0[281].val = 0x200020
scaler.SCALE_FILTER_V_BLOCK1[140].val = 0x200023
scaler.SCALE_FILTER_V_BLOCK0[282].val = 0x1c001c
scaler.SCALE_FILTER_V_BLOCK0[283].val = 0x180018
scaler.SCALE_FILTER_V_BLOCK1[141].val = 0x18001c
scaler.SCALE_FILTER_V_BLOCK0[284].val = 0x140014
scaler.SCALE_FILTER_V_BLOCK0[285].val = 0x100010
scaler.SCALE_FILTER_V_BLOCK1[142].val = 0x100014
scaler.SCALE_FILTER_V_BLOCK0[286].val = 0xb000b
scaler.SCALE_FILTER_V_BLOCK0[287].val = 0x50005
scaler.SCALE_FILTER_V_BLOCK1[143].val = 0x5000b

# DDA init H
scaler.SCALE_H_DDA_THING0 = 0
scaler.SCALE_H_DDA_THING2 = 0
scaler.SCALE_H_DDA_THING1 = 0

# horizontal scaling
scaler.SCALE_H_RATIO_0 = int(in_W / out_W * 0x400000)
scaler.SCALE_H_RATIO_4 = 0  # XXX what does this do?
scaler.SCALE_H_RATIO_1 = 0  # XXX what does this do?
scaler.SCALE_H_RATIO_2 = int(out_W / in_W * 0x400000)   # XXX what does this set do? zeroing this one out doesn't work
scaler.SCALE_H_RATIO_3 = 0  # XXX what does this set do?
scaler.SCALE_H_RATIO_5 = 0  # XXX what does this set do?
scaler.SCALE_H_FLAGS.set(EN=1)

scaler.SCALE_FILTER_H_BLOCK0[0].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[1].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[0].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[2].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[3].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[1].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[4].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[5].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[2].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[6].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[7].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[3].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[8].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[9].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[4].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[10].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[11].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[5].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[12].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[13].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[6].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[14].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[15].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[7].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[16].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[17].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[8].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[18].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[19].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[9].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[20].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[21].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[10].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[22].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[23].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[11].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[24].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[25].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[12].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[26].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[27].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[13].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[28].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[29].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[14].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[30].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[31].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[15].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[32].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[33].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[16].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[34].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[35].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[17].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[36].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[37].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[18].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[38].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[39].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[19].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[40].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[41].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[20].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[42].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[43].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[21].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[44].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[45].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[22].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[46].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[47].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[23].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[48].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[49].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[24].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[50].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[51].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[25].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[52].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[53].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[26].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[54].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[55].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[27].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[56].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[57].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[28].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[58].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[59].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[29].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[60].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[61].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[30].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[62].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[63].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[31].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[64].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[65].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[32].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[66].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[67].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[33].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[68].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[69].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[34].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[70].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[71].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[35].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[72].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[73].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[36].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[74].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[75].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[37].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[76].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[77].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[38].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[78].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[79].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[39].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[80].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[81].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[40].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[82].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[83].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[41].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[84].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[85].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[42].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[86].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[87].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[43].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[88].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[89].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[44].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[90].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[91].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[45].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[92].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[93].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[46].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[94].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[95].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[47].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[96].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[97].val = 0x50005
scaler.SCALE_FILTER_H_BLOCK1[48].val = 0x50000
scaler.SCALE_FILTER_H_BLOCK0[98].val = 0xb000b
scaler.SCALE_FILTER_H_BLOCK0[99].val = 0x100010
scaler.SCALE_FILTER_H_BLOCK1[49].val = 0x10000b
scaler.SCALE_FILTER_H_BLOCK0[100].val = 0x140014
scaler.SCALE_FILTER_H_BLOCK0[101].val = 0x180018
scaler.SCALE_FILTER_H_BLOCK1[50].val = 0x180014
scaler.SCALE_FILTER_H_BLOCK0[102].val = 0x1c001c
scaler.SCALE_FILTER_H_BLOCK0[103].val = 0x200020
scaler.SCALE_FILTER_H_BLOCK1[51].val = 0x20001c
scaler.SCALE_FILTER_H_BLOCK0[104].val = 0x230023
scaler.SCALE_FILTER_H_BLOCK0[105].val = 0x260026
scaler.SCALE_FILTER_H_BLOCK1[52].val = 0x260023
scaler.SCALE_FILTER_H_BLOCK0[106].val = 0x290029
scaler.SCALE_FILTER_H_BLOCK0[107].val = 0x2c002c
scaler.SCALE_FILTER_H_BLOCK1[53].val = 0x2c0029
scaler.SCALE_FILTER_H_BLOCK0[108].val = 0x2e002e
scaler.SCALE_FILTER_H_BLOCK0[109].val = 0x300030
scaler.SCALE_FILTER_H_BLOCK1[54].val = 0x30002e
scaler.SCALE_FILTER_H_BLOCK0[110].val = 0x320032
scaler.SCALE_FILTER_H_BLOCK0[111].val = 0x330033
scaler.SCALE_FILTER_H_BLOCK1[55].val = 0x330032
scaler.SCALE_FILTER_H_BLOCK0[112].val = 0xff87ff87
scaler.SCALE_FILTER_H_BLOCK0[113].val = 0xff90ff90
scaler.SCALE_FILTER_H_BLOCK1[56].val = 0xff90ff87
scaler.SCALE_FILTER_H_BLOCK0[114].val = 0xff99ff99
scaler.SCALE_FILTER_H_BLOCK0[115].val = 0xffa1ffa1
scaler.SCALE_FILTER_H_BLOCK1[57].val = 0xffa1ff99
scaler.SCALE_FILTER_H_BLOCK0[116].val = 0xffaaffaa
scaler.SCALE_FILTER_H_BLOCK0[117].val = 0xffb2ffb2
scaler.SCALE_FILTER_H_BLOCK1[58].val = 0xffb2ffaa
scaler.SCALE_FILTER_H_BLOCK0[118].val = 0xffbaffba
scaler.SCALE_FILTER_H_BLOCK0[119].val = 0xffc2ffc2
scaler.SCALE_FILTER_H_BLOCK1[59].val = 0xffc2ffba
scaler.SCALE_FILTER_H_BLOCK0[120].val = 0xffcaffca
scaler.SCALE_FILTER_H_BLOCK0[121].val = 0xffd2ffd2
scaler.SCALE_FILTER_H_BLOCK1[60].val = 0xffd2ffca
scaler.SCALE_FILTER_H_BLOCK0[122].val = 0xffd9ffd9
scaler.SCALE_FILTER_H_BLOCK0[123].val = 0xffe0ffe0
scaler.SCALE_FILTER_H_BLOCK1[61].val = 0xffe0ffd9
scaler.SCALE_FILTER_H_BLOCK0[124].val = 0xffe7ffe7
scaler.SCALE_FILTER_H_BLOCK0[125].val = 0xffeeffee
scaler.SCALE_FILTER_H_BLOCK1[62].val = 0xffeeffe7
scaler.SCALE_FILTER_H_BLOCK0[126].val = 0xfff4fff4
scaler.SCALE_FILTER_H_BLOCK0[127].val = 0xfffafffa
scaler.SCALE_FILTER_H_BLOCK1[63].val = 0xfffafff4
scaler.SCALE_FILTER_H_BLOCK0[128].val = 0xff06ff06
scaler.SCALE_FILTER_H_BLOCK0[129].val = 0xff0cff0c
scaler.SCALE_FILTER_H_BLOCK1[64].val = 0xff0cff06
scaler.SCALE_FILTER_H_BLOCK0[130].val = 0xff13ff13
scaler.SCALE_FILTER_H_BLOCK0[131].val = 0xff1aff1a
scaler.SCALE_FILTER_H_BLOCK1[65].val = 0xff1aff13
scaler.SCALE_FILTER_H_BLOCK0[132].val = 0xff21ff21
scaler.SCALE_FILTER_H_BLOCK0[133].val = 0xff28ff28
scaler.SCALE_FILTER_H_BLOCK1[66].val = 0xff28ff21
scaler.SCALE_FILTER_H_BLOCK0[134].val = 0xff30ff30
scaler.SCALE_FILTER_H_BLOCK0[135].val = 0xff38ff38
scaler.SCALE_FILTER_H_BLOCK1[67].val = 0xff38ff30
scaler.SCALE_FILTER_H_BLOCK0[136].val = 0xff41ff41
scaler.SCALE_FILTER_H_BLOCK0[137].val = 0xff49ff49
scaler.SCALE_FILTER_H_BLOCK1[68].val = 0xff49ff41
scaler.SCALE_FILTER_H_BLOCK0[138].val = 0xff52ff52
scaler.SCALE_FILTER_H_BLOCK0[139].val = 0xff5bff5b
scaler.SCALE_FILTER_H_BLOCK1[69].val = 0xff5bff52
scaler.SCALE_FILTER_H_BLOCK0[140].val = 0xff63ff63
scaler.SCALE_FILTER_H_BLOCK0[141].val = 0xff6cff6c
scaler.SCALE_FILTER_H_BLOCK1[70].val = 0xff6cff63
scaler.SCALE_FILTER_H_BLOCK0[142].val = 0xff75ff75
scaler.SCALE_FILTER_H_BLOCK0[143].val = 0xff7eff7e
scaler.SCALE_FILTER_H_BLOCK1[71].val = 0xff7eff75
scaler.SCALE_FILTER_H_BLOCK0[144].val = 0xff02ff02
scaler.SCALE_FILTER_H_BLOCK0[145].val = 0xfefcfefc
scaler.SCALE_FILTER_H_BLOCK1[72].val = 0xfefcff02
scaler.SCALE_FILTER_H_BLOCK0[146].val = 0xfef7fef7
scaler.SCALE_FILTER_H_BLOCK0[147].val = 0xfef3fef3
scaler.SCALE_FILTER_H_BLOCK1[73].val = 0xfef3fef7
scaler.SCALE_FILTER_H_BLOCK0[148].val = 0xfeeffeef
scaler.SCALE_FILTER_H_BLOCK0[149].val = 0xfeedfeed
scaler.SCALE_FILTER_H_BLOCK1[74].val = 0xfeedfeef
scaler.SCALE_FILTER_H_BLOCK0[150].val = 0xfeecfeec
scaler.SCALE_FILTER_H_BLOCK0[151].val = 0xfeebfeeb
scaler.SCALE_FILTER_H_BLOCK1[75].val = 0xfeebfeec
scaler.SCALE_FILTER_H_BLOCK0[152].val = 0xfeebfeeb
scaler.SCALE_FILTER_H_BLOCK0[153].val = 0xfeecfeec
scaler.SCALE_FILTER_H_BLOCK1[76].val = 0xfeecfeeb
scaler.SCALE_FILTER_H_BLOCK0[154].val = 0xfeeefeee
scaler.SCALE_FILTER_H_BLOCK0[155].val = 0xfef1fef1
scaler.SCALE_FILTER_H_BLOCK1[77].val = 0xfef1feee
scaler.SCALE_FILTER_H_BLOCK0[156].val = 0xfef4fef4
scaler.SCALE_FILTER_H_BLOCK0[157].val = 0xfef8fef8
scaler.SCALE_FILTER_H_BLOCK1[78].val = 0xfef8fef4
scaler.SCALE_FILTER_H_BLOCK0[158].val = 0xfefcfefc
scaler.SCALE_FILTER_H_BLOCK0[159].val = 0xff01ff01
scaler.SCALE_FILTER_H_BLOCK1[79].val = 0xff01fefc
scaler.SCALE_FILTER_H_BLOCK0[160].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[161].val = 0xffe7ffe7
scaler.SCALE_FILTER_H_BLOCK1[80].val = 0xffe70000
scaler.SCALE_FILTER_H_BLOCK0[162].val = 0xffcfffcf
scaler.SCALE_FILTER_H_BLOCK0[163].val = 0xffb9ffb9
scaler.SCALE_FILTER_H_BLOCK1[81].val = 0xffb9ffcf
scaler.SCALE_FILTER_H_BLOCK0[164].val = 0xffa4ffa4
scaler.SCALE_FILTER_H_BLOCK0[165].val = 0xff90ff90
scaler.SCALE_FILTER_H_BLOCK1[82].val = 0xff90ffa4
scaler.SCALE_FILTER_H_BLOCK0[166].val = 0xff7dff7d
scaler.SCALE_FILTER_H_BLOCK0[167].val = 0xff6bff6b
scaler.SCALE_FILTER_H_BLOCK1[83].val = 0xff6bff7d
scaler.SCALE_FILTER_H_BLOCK0[168].val = 0xff5bff5b
scaler.SCALE_FILTER_H_BLOCK0[169].val = 0xff4cff4c
scaler.SCALE_FILTER_H_BLOCK1[84].val = 0xff4cff5b
scaler.SCALE_FILTER_H_BLOCK0[170].val = 0xff3eff3e
scaler.SCALE_FILTER_H_BLOCK0[171].val = 0xff31ff31
scaler.SCALE_FILTER_H_BLOCK1[85].val = 0xff31ff3e
scaler.SCALE_FILTER_H_BLOCK0[172].val = 0xff26ff26
scaler.SCALE_FILTER_H_BLOCK0[173].val = 0xff1bff1b
scaler.SCALE_FILTER_H_BLOCK1[86].val = 0xff1bff26
scaler.SCALE_FILTER_H_BLOCK0[174].val = 0xff12ff12
scaler.SCALE_FILTER_H_BLOCK0[175].val = 0xff0aff0a
scaler.SCALE_FILTER_H_BLOCK1[87].val = 0xff0aff12
scaler.SCALE_FILTER_H_BLOCK0[176].val = 0x2210221
scaler.SCALE_FILTER_H_BLOCK0[177].val = 0x1f901f9
scaler.SCALE_FILTER_H_BLOCK1[88].val = 0x1f90221
scaler.SCALE_FILTER_H_BLOCK0[178].val = 0x1d001d0
scaler.SCALE_FILTER_H_BLOCK0[179].val = 0x1a901a9
scaler.SCALE_FILTER_H_BLOCK1[89].val = 0x1a901d0
scaler.SCALE_FILTER_H_BLOCK0[180].val = 0x1820182
scaler.SCALE_FILTER_H_BLOCK0[181].val = 0x15d015d
scaler.SCALE_FILTER_H_BLOCK1[90].val = 0x15d0182
scaler.SCALE_FILTER_H_BLOCK0[182].val = 0x1380138
scaler.SCALE_FILTER_H_BLOCK0[183].val = 0x1140114
scaler.SCALE_FILTER_H_BLOCK1[91].val = 0x1140138
scaler.SCALE_FILTER_H_BLOCK0[184].val = 0xf100f1
scaler.SCALE_FILTER_H_BLOCK0[185].val = 0xcf00cf
scaler.SCALE_FILTER_H_BLOCK1[92].val = 0xcf00f1
scaler.SCALE_FILTER_H_BLOCK0[186].val = 0xae00ae
scaler.SCALE_FILTER_H_BLOCK0[187].val = 0x8e008e
scaler.SCALE_FILTER_H_BLOCK1[93].val = 0x8e00ae
scaler.SCALE_FILTER_H_BLOCK0[188].val = 0x6f006f
scaler.SCALE_FILTER_H_BLOCK0[189].val = 0x520052
scaler.SCALE_FILTER_H_BLOCK1[94].val = 0x52006f
scaler.SCALE_FILTER_H_BLOCK0[190].val = 0x350035
scaler.SCALE_FILTER_H_BLOCK0[191].val = 0x1a001a
scaler.SCALE_FILTER_H_BLOCK1[95].val = 0x1a0035
scaler.SCALE_FILTER_H_BLOCK0[192].val = 0x4e404e4
scaler.SCALE_FILTER_H_BLOCK0[193].val = 0x4b804b8
scaler.SCALE_FILTER_H_BLOCK1[96].val = 0x4b804e4
scaler.SCALE_FILTER_H_BLOCK0[194].val = 0x48b048b
scaler.SCALE_FILTER_H_BLOCK0[195].val = 0x45f045f
scaler.SCALE_FILTER_H_BLOCK1[97].val = 0x45f048b
scaler.SCALE_FILTER_H_BLOCK0[196].val = 0x4320432
scaler.SCALE_FILTER_H_BLOCK0[197].val = 0x4050405
scaler.SCALE_FILTER_H_BLOCK1[98].val = 0x4050432
scaler.SCALE_FILTER_H_BLOCK0[198].val = 0x3d803d8
scaler.SCALE_FILTER_H_BLOCK0[199].val = 0x3ab03ab
scaler.SCALE_FILTER_H_BLOCK1[99].val = 0x3ab03d8
scaler.SCALE_FILTER_H_BLOCK0[200].val = 0x37e037e
scaler.SCALE_FILTER_H_BLOCK0[201].val = 0x3510351
scaler.SCALE_FILTER_H_BLOCK1[100].val = 0x351037e
scaler.SCALE_FILTER_H_BLOCK0[202].val = 0x3240324
scaler.SCALE_FILTER_H_BLOCK0[203].val = 0x2f802f8
scaler.SCALE_FILTER_H_BLOCK1[101].val = 0x2f80324
scaler.SCALE_FILTER_H_BLOCK0[204].val = 0x2cc02cc
scaler.SCALE_FILTER_H_BLOCK0[205].val = 0x2a102a1
scaler.SCALE_FILTER_H_BLOCK1[102].val = 0x2a102cc
scaler.SCALE_FILTER_H_BLOCK0[206].val = 0x2760276
scaler.SCALE_FILTER_H_BLOCK0[207].val = 0x24b024b
scaler.SCALE_FILTER_H_BLOCK1[103].val = 0x24b0276
scaler.SCALE_FILTER_H_BLOCK0[208].val = 0x73b073b
scaler.SCALE_FILTER_H_BLOCK0[209].val = 0x71e071e
scaler.SCALE_FILTER_H_BLOCK1[104].val = 0x71e073b
scaler.SCALE_FILTER_H_BLOCK0[210].val = 0x7000700
scaler.SCALE_FILTER_H_BLOCK0[211].val = 0x6e106e1
scaler.SCALE_FILTER_H_BLOCK1[105].val = 0x6e10700
scaler.SCALE_FILTER_H_BLOCK0[212].val = 0x6c006c0
scaler.SCALE_FILTER_H_BLOCK0[213].val = 0x69e069e
scaler.SCALE_FILTER_H_BLOCK1[106].val = 0x69e06c0
scaler.SCALE_FILTER_H_BLOCK0[214].val = 0x67a067a
scaler.SCALE_FILTER_H_BLOCK0[215].val = 0x6560656
scaler.SCALE_FILTER_H_BLOCK1[107].val = 0x656067a
scaler.SCALE_FILTER_H_BLOCK0[216].val = 0x6300630
scaler.SCALE_FILTER_H_BLOCK0[217].val = 0x6090609
scaler.SCALE_FILTER_H_BLOCK1[108].val = 0x6090630
scaler.SCALE_FILTER_H_BLOCK0[218].val = 0x5e205e2
scaler.SCALE_FILTER_H_BLOCK0[219].val = 0x5b905b9
scaler.SCALE_FILTER_H_BLOCK1[109].val = 0x5b905e2
scaler.SCALE_FILTER_H_BLOCK0[220].val = 0x5900590
scaler.SCALE_FILTER_H_BLOCK0[221].val = 0x5660566
scaler.SCALE_FILTER_H_BLOCK1[110].val = 0x5660590
scaler.SCALE_FILTER_H_BLOCK0[222].val = 0x53b053b
scaler.SCALE_FILTER_H_BLOCK0[223].val = 0x5100510
scaler.SCALE_FILTER_H_BLOCK1[111].val = 0x510053b
scaler.SCALE_FILTER_H_BLOCK0[224].val = 0x82c082c
scaler.SCALE_FILTER_H_BLOCK0[225].val = 0x82b082b
scaler.SCALE_FILTER_H_BLOCK1[112].val = 0x82b082c
scaler.SCALE_FILTER_H_BLOCK0[226].val = 0x8280828
scaler.SCALE_FILTER_H_BLOCK0[227].val = 0x8200820
scaler.SCALE_FILTER_H_BLOCK1[113].val = 0x8200828
scaler.SCALE_FILTER_H_BLOCK0[228].val = 0x81b081b
scaler.SCALE_FILTER_H_BLOCK0[229].val = 0x8130813
scaler.SCALE_FILTER_H_BLOCK1[114].val = 0x813081b
scaler.SCALE_FILTER_H_BLOCK0[230].val = 0x8080808
scaler.SCALE_FILTER_H_BLOCK0[231].val = 0x7fc07fc
scaler.SCALE_FILTER_H_BLOCK1[115].val = 0x7fc0808
scaler.SCALE_FILTER_H_BLOCK0[232].val = 0x7ed07ed
scaler.SCALE_FILTER_H_BLOCK0[233].val = 0x7dd07dd
scaler.SCALE_FILTER_H_BLOCK1[116].val = 0x7dd07ed
scaler.SCALE_FILTER_H_BLOCK0[234].val = 0x7cb07cb
scaler.SCALE_FILTER_H_BLOCK0[235].val = 0x7b607b6
scaler.SCALE_FILTER_H_BLOCK1[117].val = 0x7b607cb
scaler.SCALE_FILTER_H_BLOCK0[236].val = 0x7a207a2
scaler.SCALE_FILTER_H_BLOCK0[237].val = 0x78a078a
scaler.SCALE_FILTER_H_BLOCK1[118].val = 0x78a07a2
scaler.SCALE_FILTER_H_BLOCK0[238].val = 0x7710771
scaler.SCALE_FILTER_H_BLOCK0[239].val = 0x7570757
scaler.SCALE_FILTER_H_BLOCK1[119].val = 0x7570771
scaler.SCALE_FILTER_H_BLOCK0[240].val = 0x73d073d
scaler.SCALE_FILTER_H_BLOCK0[241].val = 0x7570757
scaler.SCALE_FILTER_H_BLOCK1[120].val = 0x757073d
scaler.SCALE_FILTER_H_BLOCK0[242].val = 0x7710771
scaler.SCALE_FILTER_H_BLOCK0[243].val = 0x78a078a
scaler.SCALE_FILTER_H_BLOCK1[121].val = 0x78a0771
scaler.SCALE_FILTER_H_BLOCK0[244].val = 0x7a207a2
scaler.SCALE_FILTER_H_BLOCK0[245].val = 0x7b607b6
scaler.SCALE_FILTER_H_BLOCK1[122].val = 0x7b607a2
scaler.SCALE_FILTER_H_BLOCK0[246].val = 0x7cb07cb
scaler.SCALE_FILTER_H_BLOCK0[247].val = 0x7dd07dd
scaler.SCALE_FILTER_H_BLOCK1[123].val = 0x7dd07cb
scaler.SCALE_FILTER_H_BLOCK0[248].val = 0x7ed07ed
scaler.SCALE_FILTER_H_BLOCK0[249].val = 0x7fc07fc
scaler.SCALE_FILTER_H_BLOCK1[124].val = 0x7fc07ed
scaler.SCALE_FILTER_H_BLOCK0[250].val = 0x8080808
scaler.SCALE_FILTER_H_BLOCK0[251].val = 0x8130813
scaler.SCALE_FILTER_H_BLOCK1[125].val = 0x8130808
scaler.SCALE_FILTER_H_BLOCK0[252].val = 0x81b081b
scaler.SCALE_FILTER_H_BLOCK0[253].val = 0x8200820
scaler.SCALE_FILTER_H_BLOCK1[126].val = 0x820081b
scaler.SCALE_FILTER_H_BLOCK0[254].val = 0x8280828
scaler.SCALE_FILTER_H_BLOCK0[255].val = 0x82b082b
scaler.SCALE_FILTER_H_BLOCK1[127].val = 0x82b0828
scaler.SCALE_FILTER_H_BLOCK0[256].val = 0x4e404e4
scaler.SCALE_FILTER_H_BLOCK0[257].val = 0x5100510
scaler.SCALE_FILTER_H_BLOCK1[128].val = 0x51004e4
scaler.SCALE_FILTER_H_BLOCK0[258].val = 0x53b053b
scaler.SCALE_FILTER_H_BLOCK0[259].val = 0x5660566
scaler.SCALE_FILTER_H_BLOCK1[129].val = 0x566053b
scaler.SCALE_FILTER_H_BLOCK0[260].val = 0x5900590
scaler.SCALE_FILTER_H_BLOCK0[261].val = 0x5b905b9
scaler.SCALE_FILTER_H_BLOCK1[130].val = 0x5b90590
scaler.SCALE_FILTER_H_BLOCK0[262].val = 0x5e205e2
scaler.SCALE_FILTER_H_BLOCK0[263].val = 0x6090609
scaler.SCALE_FILTER_H_BLOCK1[131].val = 0x60905e2
scaler.SCALE_FILTER_H_BLOCK0[264].val = 0x6300630
scaler.SCALE_FILTER_H_BLOCK0[265].val = 0x6560656
scaler.SCALE_FILTER_H_BLOCK1[132].val = 0x6560630
scaler.SCALE_FILTER_H_BLOCK0[266].val = 0x67a067a
scaler.SCALE_FILTER_H_BLOCK0[267].val = 0x69e069e
scaler.SCALE_FILTER_H_BLOCK1[133].val = 0x69e067a
scaler.SCALE_FILTER_H_BLOCK0[268].val = 0x6c006c0
scaler.SCALE_FILTER_H_BLOCK0[269].val = 0x6e106e1
scaler.SCALE_FILTER_H_BLOCK1[134].val = 0x6e106c0
scaler.SCALE_FILTER_H_BLOCK0[270].val = 0x7000700
scaler.SCALE_FILTER_H_BLOCK0[271].val = 0x71e071e
scaler.SCALE_FILTER_H_BLOCK1[135].val = 0x71e0700
scaler.SCALE_FILTER_H_BLOCK0[272].val = 0x2210221
scaler.SCALE_FILTER_H_BLOCK0[273].val = 0x24b024b
scaler.SCALE_FILTER_H_BLOCK1[136].val = 0x24b0221
scaler.SCALE_FILTER_H_BLOCK0[274].val = 0x2760276
scaler.SCALE_FILTER_H_BLOCK0[275].val = 0x2a102a1
scaler.SCALE_FILTER_H_BLOCK1[137].val = 0x2a10276
scaler.SCALE_FILTER_H_BLOCK0[276].val = 0x2cc02cc
scaler.SCALE_FILTER_H_BLOCK0[277].val = 0x2f802f8
scaler.SCALE_FILTER_H_BLOCK1[138].val = 0x2f802cc
scaler.SCALE_FILTER_H_BLOCK0[278].val = 0x3240324
scaler.SCALE_FILTER_H_BLOCK0[279].val = 0x3510351
scaler.SCALE_FILTER_H_BLOCK1[139].val = 0x3510324
scaler.SCALE_FILTER_H_BLOCK0[280].val = 0x37e037e
scaler.SCALE_FILTER_H_BLOCK0[281].val = 0x3ab03ab
scaler.SCALE_FILTER_H_BLOCK1[140].val = 0x3ab037e
scaler.SCALE_FILTER_H_BLOCK0[282].val = 0x3d803d8
scaler.SCALE_FILTER_H_BLOCK0[283].val = 0x4050405
scaler.SCALE_FILTER_H_BLOCK1[141].val = 0x40503d8
scaler.SCALE_FILTER_H_BLOCK0[284].val = 0x4320432
scaler.SCALE_FILTER_H_BLOCK0[285].val = 0x45f045f
scaler.SCALE_FILTER_H_BLOCK1[142].val = 0x45f0432
scaler.SCALE_FILTER_H_BLOCK0[286].val = 0x48b048b
scaler.SCALE_FILTER_H_BLOCK0[287].val = 0x4b804b8
scaler.SCALE_FILTER_H_BLOCK1[143].val = 0x4b8048b
scaler.SCALE_FILTER_H_BLOCK0[288].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[289].val = 0x1a001a
scaler.SCALE_FILTER_H_BLOCK1[144].val = 0x1a0000
scaler.SCALE_FILTER_H_BLOCK0[290].val = 0x350035
scaler.SCALE_FILTER_H_BLOCK0[291].val = 0x520052
scaler.SCALE_FILTER_H_BLOCK1[145].val = 0x520035
scaler.SCALE_FILTER_H_BLOCK0[292].val = 0x6f006f
scaler.SCALE_FILTER_H_BLOCK0[293].val = 0x8e008e
scaler.SCALE_FILTER_H_BLOCK1[146].val = 0x8e006f
scaler.SCALE_FILTER_H_BLOCK0[294].val = 0xae00ae
scaler.SCALE_FILTER_H_BLOCK0[295].val = 0xcf00cf
scaler.SCALE_FILTER_H_BLOCK1[147].val = 0xcf00ae
scaler.SCALE_FILTER_H_BLOCK0[296].val = 0xf100f1
scaler.SCALE_FILTER_H_BLOCK0[297].val = 0x1140114
scaler.SCALE_FILTER_H_BLOCK1[148].val = 0x11400f1
scaler.SCALE_FILTER_H_BLOCK0[298].val = 0x1380138
scaler.SCALE_FILTER_H_BLOCK0[299].val = 0x15d015d
scaler.SCALE_FILTER_H_BLOCK1[149].val = 0x15d0138
scaler.SCALE_FILTER_H_BLOCK0[300].val = 0x1820182
scaler.SCALE_FILTER_H_BLOCK0[301].val = 0x1a901a9
scaler.SCALE_FILTER_H_BLOCK1[150].val = 0x1a90182
scaler.SCALE_FILTER_H_BLOCK0[302].val = 0x1d001d0
scaler.SCALE_FILTER_H_BLOCK0[303].val = 0x1f901f9
scaler.SCALE_FILTER_H_BLOCK1[151].val = 0x1f901d0
scaler.SCALE_FILTER_H_BLOCK0[304].val = 0xff02ff02
scaler.SCALE_FILTER_H_BLOCK0[305].val = 0xff0aff0a
scaler.SCALE_FILTER_H_BLOCK1[152].val = 0xff0aff02
scaler.SCALE_FILTER_H_BLOCK0[306].val = 0xff12ff12
scaler.SCALE_FILTER_H_BLOCK0[307].val = 0xff1bff1b
scaler.SCALE_FILTER_H_BLOCK1[153].val = 0xff1bff12
scaler.SCALE_FILTER_H_BLOCK0[308].val = 0xff26ff26
scaler.SCALE_FILTER_H_BLOCK0[309].val = 0xff31ff31
scaler.SCALE_FILTER_H_BLOCK1[154].val = 0xff31ff26
scaler.SCALE_FILTER_H_BLOCK0[310].val = 0xff3eff3e
scaler.SCALE_FILTER_H_BLOCK0[311].val = 0xff4cff4c
scaler.SCALE_FILTER_H_BLOCK1[155].val = 0xff4cff3e
scaler.SCALE_FILTER_H_BLOCK0[312].val = 0xff5bff5b
scaler.SCALE_FILTER_H_BLOCK0[313].val = 0xff6bff6b
scaler.SCALE_FILTER_H_BLOCK1[156].val = 0xff6bff5b
scaler.SCALE_FILTER_H_BLOCK0[314].val = 0xff7dff7d
scaler.SCALE_FILTER_H_BLOCK0[315].val = 0xff90ff90
scaler.SCALE_FILTER_H_BLOCK1[157].val = 0xff90ff7d
scaler.SCALE_FILTER_H_BLOCK0[316].val = 0xffa4ffa4
scaler.SCALE_FILTER_H_BLOCK0[317].val = 0xffb9ffb9
scaler.SCALE_FILTER_H_BLOCK1[158].val = 0xffb9ffa4
scaler.SCALE_FILTER_H_BLOCK0[318].val = 0xffcfffcf
scaler.SCALE_FILTER_H_BLOCK0[319].val = 0xffe7ffe7
scaler.SCALE_FILTER_H_BLOCK1[159].val = 0xffe7ffcf
scaler.SCALE_FILTER_H_BLOCK0[320].val = 0xff06ff06
scaler.SCALE_FILTER_H_BLOCK0[321].val = 0xff01ff01
scaler.SCALE_FILTER_H_BLOCK1[160].val = 0xff01ff06
scaler.SCALE_FILTER_H_BLOCK0[322].val = 0xfefcfefc
scaler.SCALE_FILTER_H_BLOCK0[323].val = 0xfef8fef8
scaler.SCALE_FILTER_H_BLOCK1[161].val = 0xfef8fefc
scaler.SCALE_FILTER_H_BLOCK0[324].val = 0xfef4fef4
scaler.SCALE_FILTER_H_BLOCK0[325].val = 0xfef1fef1
scaler.SCALE_FILTER_H_BLOCK1[162].val = 0xfef1fef4
scaler.SCALE_FILTER_H_BLOCK0[326].val = 0xfeeefeee
scaler.SCALE_FILTER_H_BLOCK0[327].val = 0xfeecfeec
scaler.SCALE_FILTER_H_BLOCK1[163].val = 0xfeecfeee
scaler.SCALE_FILTER_H_BLOCK0[328].val = 0xfeebfeeb
scaler.SCALE_FILTER_H_BLOCK0[329].val = 0xfeebfeeb
scaler.SCALE_FILTER_H_BLOCK1[164].val = 0xfeebfeeb
scaler.SCALE_FILTER_H_BLOCK0[330].val = 0xfeecfeec
scaler.SCALE_FILTER_H_BLOCK0[331].val = 0xfeedfeed
scaler.SCALE_FILTER_H_BLOCK1[165].val = 0xfeedfeec
scaler.SCALE_FILTER_H_BLOCK0[332].val = 0xfeeffeef
scaler.SCALE_FILTER_H_BLOCK0[333].val = 0xfef3fef3
scaler.SCALE_FILTER_H_BLOCK1[166].val = 0xfef3feef
scaler.SCALE_FILTER_H_BLOCK0[334].val = 0xfef7fef7
scaler.SCALE_FILTER_H_BLOCK0[335].val = 0xfefcfefc
scaler.SCALE_FILTER_H_BLOCK1[167].val = 0xfefcfef7
scaler.SCALE_FILTER_H_BLOCK0[336].val = 0xff87ff87
scaler.SCALE_FILTER_H_BLOCK0[337].val = 0xff7eff7e
scaler.SCALE_FILTER_H_BLOCK1[168].val = 0xff7eff87
scaler.SCALE_FILTER_H_BLOCK0[338].val = 0xff75ff75
scaler.SCALE_FILTER_H_BLOCK0[339].val = 0xff6cff6c
scaler.SCALE_FILTER_H_BLOCK1[169].val = 0xff6cff75
scaler.SCALE_FILTER_H_BLOCK0[340].val = 0xff63ff63
scaler.SCALE_FILTER_H_BLOCK0[341].val = 0xff5bff5b
scaler.SCALE_FILTER_H_BLOCK1[170].val = 0xff5bff63
scaler.SCALE_FILTER_H_BLOCK0[342].val = 0xff52ff52
scaler.SCALE_FILTER_H_BLOCK0[343].val = 0xff49ff49
scaler.SCALE_FILTER_H_BLOCK1[171].val = 0xff49ff52
scaler.SCALE_FILTER_H_BLOCK0[344].val = 0xff41ff41
scaler.SCALE_FILTER_H_BLOCK0[345].val = 0xff38ff38
scaler.SCALE_FILTER_H_BLOCK1[172].val = 0xff38ff41
scaler.SCALE_FILTER_H_BLOCK0[346].val = 0xff30ff30
scaler.SCALE_FILTER_H_BLOCK0[347].val = 0xff28ff28
scaler.SCALE_FILTER_H_BLOCK1[173].val = 0xff28ff30
scaler.SCALE_FILTER_H_BLOCK0[348].val = 0xff21ff21
scaler.SCALE_FILTER_H_BLOCK0[349].val = 0xff1aff1a
scaler.SCALE_FILTER_H_BLOCK1[174].val = 0xff1aff21
scaler.SCALE_FILTER_H_BLOCK0[350].val = 0xff13ff13
scaler.SCALE_FILTER_H_BLOCK0[351].val = 0xff0cff0c
scaler.SCALE_FILTER_H_BLOCK1[175].val = 0xff0cff13
scaler.SCALE_FILTER_H_BLOCK0[352].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[353].val = 0xfffafffa
scaler.SCALE_FILTER_H_BLOCK1[176].val = 0xfffa0000
scaler.SCALE_FILTER_H_BLOCK0[354].val = 0xfff4fff4
scaler.SCALE_FILTER_H_BLOCK0[355].val = 0xffeeffee
scaler.SCALE_FILTER_H_BLOCK1[177].val = 0xffeefff4
scaler.SCALE_FILTER_H_BLOCK0[356].val = 0xffe7ffe7
scaler.SCALE_FILTER_H_BLOCK0[357].val = 0xffe0ffe0
scaler.SCALE_FILTER_H_BLOCK1[178].val = 0xffe0ffe7
scaler.SCALE_FILTER_H_BLOCK0[358].val = 0xffd9ffd9
scaler.SCALE_FILTER_H_BLOCK0[359].val = 0xffd2ffd2
scaler.SCALE_FILTER_H_BLOCK1[179].val = 0xffd2ffd9
scaler.SCALE_FILTER_H_BLOCK0[360].val = 0xffcaffca
scaler.SCALE_FILTER_H_BLOCK0[361].val = 0xffc2ffc2
scaler.SCALE_FILTER_H_BLOCK1[180].val = 0xffc2ffca
scaler.SCALE_FILTER_H_BLOCK0[362].val = 0xffbaffba
scaler.SCALE_FILTER_H_BLOCK0[363].val = 0xffb2ffb2
scaler.SCALE_FILTER_H_BLOCK1[181].val = 0xffb2ffba
scaler.SCALE_FILTER_H_BLOCK0[364].val = 0xffaaffaa
scaler.SCALE_FILTER_H_BLOCK0[365].val = 0xffa1ffa1
scaler.SCALE_FILTER_H_BLOCK1[182].val = 0xffa1ffaa
scaler.SCALE_FILTER_H_BLOCK0[366].val = 0xff99ff99
scaler.SCALE_FILTER_H_BLOCK0[367].val = 0xff90ff90
scaler.SCALE_FILTER_H_BLOCK1[183].val = 0xff90ff99
scaler.SCALE_FILTER_H_BLOCK0[368].val = 0x340034
scaler.SCALE_FILTER_H_BLOCK0[369].val = 0x330033
scaler.SCALE_FILTER_H_BLOCK1[184].val = 0x330034
scaler.SCALE_FILTER_H_BLOCK0[370].val = 0x320032
scaler.SCALE_FILTER_H_BLOCK0[371].val = 0x300030
scaler.SCALE_FILTER_H_BLOCK1[185].val = 0x300032
scaler.SCALE_FILTER_H_BLOCK0[372].val = 0x2e002e
scaler.SCALE_FILTER_H_BLOCK0[373].val = 0x2c002c
scaler.SCALE_FILTER_H_BLOCK1[186].val = 0x2c002e
scaler.SCALE_FILTER_H_BLOCK0[374].val = 0x290029
scaler.SCALE_FILTER_H_BLOCK0[375].val = 0x260026
scaler.SCALE_FILTER_H_BLOCK1[187].val = 0x260029
scaler.SCALE_FILTER_H_BLOCK0[376].val = 0x230023
scaler.SCALE_FILTER_H_BLOCK0[377].val = 0x200020
scaler.SCALE_FILTER_H_BLOCK1[188].val = 0x200023
scaler.SCALE_FILTER_H_BLOCK0[378].val = 0x1c001c
scaler.SCALE_FILTER_H_BLOCK0[379].val = 0x180018
scaler.SCALE_FILTER_H_BLOCK1[189].val = 0x18001c
scaler.SCALE_FILTER_H_BLOCK0[380].val = 0x140014
scaler.SCALE_FILTER_H_BLOCK0[381].val = 0x100010
scaler.SCALE_FILTER_H_BLOCK1[190].val = 0x100014
scaler.SCALE_FILTER_H_BLOCK0[382].val = 0xb000b
scaler.SCALE_FILTER_H_BLOCK0[383].val = 0x50005
scaler.SCALE_FILTER_H_BLOCK1[191].val = 0x5000b
scaler.SCALE_FILTER_H_BLOCK0[384].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[385].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[192].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[386].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[387].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[193].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[388].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[389].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[194].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[390].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[391].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[195].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[392].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[393].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[196].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[394].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[395].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[197].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[396].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[397].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[198].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[398].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[399].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[199].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[400].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[401].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[200].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[402].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[403].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[201].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[404].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[405].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[202].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[406].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[407].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[203].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[408].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[409].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[204].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[410].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[411].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[205].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[412].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[413].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[206].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[414].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[415].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[207].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[416].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[417].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[208].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[418].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[419].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[209].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[420].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[421].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[210].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[422].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[423].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[211].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[424].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[425].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[212].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[426].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[427].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[213].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[428].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[429].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[214].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[430].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[431].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[215].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[432].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[433].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[216].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[434].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[435].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[217].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[436].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[437].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[218].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[438].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[439].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[219].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[440].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[441].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[220].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[442].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[443].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[221].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[444].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[445].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[222].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[446].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[447].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[223].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[448].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[449].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[224].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[450].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[451].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[225].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[452].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[453].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[226].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[454].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[455].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[227].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[456].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[457].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[228].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[458].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[459].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[229].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[460].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[461].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[230].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[462].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[463].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[231].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[464].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[465].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[232].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[466].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[467].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[233].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[468].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[469].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[234].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[470].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[471].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[235].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[472].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[473].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[236].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[474].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[475].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[237].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[476].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[477].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[238].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[478].val = 0x0
scaler.SCALE_FILTER_H_BLOCK0[479].val = 0x0
scaler.SCALE_FILTER_H_BLOCK1[239].val = 0x0

# pseudo linear scaling
scaler.PSEUDO_LINEAR_SCALING = 0

# reshape
p.write32(scaler_base + 0xe8, 0x0)

# reset CM 3x
p.write32(scaler_base + 0x3800, 0x0)
p.write32(scaler_base + 0x3800, 0x0)
p.write32(scaler_base + 0x3800, 0x0)

# enable prescaler
p.write32(scaler_base + 0x824, 0xc)

# alpha override
p.write32(scaler_base + 0x8c, 0xffff)

# dither
p.write32(scaler_base + 0xa00, 0x0)

# commit convert
p.write32(scaler_base + 0x13808, 0x0)
p.write32(scaler_base + 0x1380c, 0x0)
p.write32(scaler_base + 0x13800, 0x0)
p.write32(scaler_base + 0x13804, 0x0)

# convert map
p.write32(scaler_base + 0x13810, 0x8)
p.write32(scaler_base + 0x13814, 0x8)
p.write32(scaler_base + 0x13818, 0x8)
p.write32(scaler_base + 0x1381c, 0x8)
p.write32(scaler_base + 0x13804, 0x0)
p.write32(scaler_base + 0x13c04, 0x0)
p.write32(scaler_base + 0x13c10, 0x8)
p.write32(scaler_base + 0x13c14, 0x8)
p.write32(scaler_base + 0x13c18, 0x8)
p.write32(scaler_base + 0x13c1c, 0x8)

# commit revert
p.write32(scaler_base + 0x13c00, 0x1)

# (don't) program histogram
p.write32(scaler_base + 0x3000, 0x0)
p.write32(scaler_base + 0x124, 0x0)

# tag transform registers
p.write32(scaler_base + 0x110, 0x1)

# start
p.write32(scaler_base + 0x98, 0xfffffffe)
scaler.START = 1

start_time = time.time()
while scaler.MSR_GLBL_IRQSTS.reg.DONE == 0:
    if time.time() - start_time > 5:
        print("TIMED OUT!!!")
        break

print(f"IRQ status is now {scaler.MSR_GLBL_IRQSTS}")
print(f"Debug status is now {scaler.MSR_CTRL_DBGSTS}")

out_buf_new = iface.readmem(out_buf_phys, out_SZ)
chexdump(out_buf_new)

with Image.new(mode='RGBA', size=(out_W, out_H)) as im:
    for y in range(out_H):
        for x in range(out_W):
            block = out_buf_new[
                y*out_STRIDE + x*out_BYTESPP:
                y*out_STRIDE + (x+1)*out_BYTESPP]

            r, g, b, a = block
            im.putpixel((x, y), (r, g, b, a))

    im.save(output_image_fn)



import json, sys
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("regfile")
args = parser.parse_args()

data = json.load(open(args.regfile))
for reg in data:
    name = reg['name']

    if name[-4:-1] == "_EL":
        name = name[:-4]

    if not reg.get("fieldsets", []):
        continue

    print(f"# {reg['name']}")
    print(f"class {name}(Register64):")

    for fieldset in reg.get("fieldsets", []):
        if "instance" in fieldset:
            print(f"# {fieldset['instance']}")
        for f in fieldset["fields"]:
            fname = f["name"]
            msb, lsb = f["msb"], f["lsb"]

            if msb == lsb:
                print(f"    {fname} = {lsb}")
            else:
                print(f"    {fname} = {msb}, {lsb}")

    print()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, time, random, array
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1 import asm

PAGE_SIZE = 16384

REPETITIONS = 1000

TEST_ECORE = 1
TEST_PCORE = 5

S = 8
ITERS = 1024
SIZE_DATA_ARRAY = ITERS * S * 4

WRITE_MAX = SIZE_DATA_ARRAY // 4

seq_data_buf = u.memalign(PAGE_SIZE, SIZE_DATA_ARRAY)
random_data_buf = u.memalign(PAGE_SIZE, SIZE_DATA_ARRAY)

seq_data = [0] * WRITE_MAX
for i in range(0, WRITE_MAX, S):
    seq_data[i] = i + S

random_data = [0] * WRITE_MAX
order = list(range(1, ITERS))
random.shuffle(order)
off = 0
for i in range(ITERS - 1):
    next_off = order[i]
    assert random_data[S * off] == 0
    random_data[S * off] = S * next_off
    off = next_off

print(f"Seq data buf: {seq_data_buf:#x}")
print(f"Random data buf: {random_data_buf:#x}")

iface.writemem(seq_data_buf, array.array('I', seq_data).tobytes())
iface.writemem(random_data_buf, array.array('I', random_data).tobytes())

freq = u.mrs(CNTFRQ_EL0)
code = u.malloc(0x1000)

util = asm.ARMAsm(f"""
test:
    mov x6, x2
    mov x5, x0
    mov x7, #0
    mov x4, #0
    mov x2, x1
    mov x3, #0
1:
    ldr w3, [x5, x3, lsl #2]
    subs x2, x2, #1
    b.ne 1b

3:
    mov x2, x1
    mov x3, #0

    dsb sy
    isb
    mrs x9, S3_2_c15_c0_0
    isb

2:
    ldr w3, [x5, x3, lsl #2]
    subs x2, x2, #1
    b.ne 2b

    dsb sy
    isb
    mrs x10, S3_2_c15_c0_0
    isb

    sub x0, x10, x9
    add x7, x7, x0

    subs x6, x6, #1
    b.ne 3b

    mov x0, x7
    ret
""", code)
iface.writemem(code, util.data)
p.dc_cvau(code, len(util.data))
p.ic_ivau(code, len(util.data))

# Set higher cpufreq pstate on all clusters
p.cpufreq_init()

p.smp_start_secondaries()
p.smp_set_wfe_mode(False)

def cpu_call(cpu, x, *args):
    return p.smp_call_sync(cpu, x | REGION_RX_EL1, *args)

def init_core(cpu):
    p.mmu_init_secondary(cpu)

    def mrs(x):
        return u.mrs(x, call=lambda x, *args: cpu_call(cpu, x, *args))
    def msr(x, v):
        u.msr(x, v, call=lambda x, *args: cpu_call(cpu, x, *args))

    msr(SPRR_CONFIG_EL1, 1)

    is_ecore = not (mrs(MPIDR_EL1) & (1 << 16))
    # Enable DC MVA ops
    v = mrs(EHID4_EL1 if is_ecore else HID4_EL1)
    v &= ~(1 << 11)
    msr(EHID4_EL1 if is_ecore else HID4_EL1, v)

    # Enable PMU
    v = mrs(PMCR0_EL1)
    v |= 1 | (1<<30)
    msr(PMCR0_EL1, v)
    msr(PMCR1_EL1, 0xffffffffffffffff)

    v = mrs(CNTKCTL_EL1)
    v |= 3
    msr(CNTKCTL_EL1, v)

    # Enable user cache ops
    v = mrs(SCTLR_EL1)
    v |= (1 << 26)
    msr(SCTLR_EL1, v)

def cpu_msr(cpu, x, v):
    u.msr(x, v, call=lambda x, *args: cpu_call(cpu, x, *args))

init_core(TEST_ECORE)
init_core(TEST_PCORE)

# Enable DC MVA ops
v = u.mrs(EHID4_EL1)
v &= ~(1 << 11)
u.msr(EHID4_EL1, v)

def test_cpu(cpu, buf, iters):
    elapsed = p.smp_call_sync(cpu, util.test | REGION_RX_EL1, buf, iters, REPETITIONS)
    return elapsed / iters / REPETITIONS

def run_tests():
    a = test_cpu(TEST_ECORE, seq_data_buf, ITERS)
    b = test_cpu(TEST_ECORE, random_data_buf, ITERS)
    c = test_cpu(TEST_PCORE, seq_data_buf, ITERS)
    d = test_cpu(TEST_PCORE, random_data_buf, ITERS)
    print(f"    ECore seq: {a:.02f}, ECore random: {b:.02f}")
    print(f"    PCore seq: {c:.02f}, PCore random: {d:.02f}")

print("Testing with SSBS=1 (load/store speculation permitted)")
for cpu in (TEST_ECORE, TEST_PCORE):
    cpu_msr(cpu, SSBS, 1<<12)

run_tests()
print("Testing with SSBS=0 (load/store speculation disallowed)")
for cpu in (TEST_ECORE, TEST_PCORE):
    cpu_msr(cpu, SSBS, 0)

run_tests()



# SPDX-License-Identifier: MIT
from ..utils import *

class ISP_REVISION(Register32):
    REVISION = 15, 0

class ISPRegs(RegMap):
    ISP_PMGR_0 = 0x738, Register32
    ISP_PMGR_1 = 0x798, Register32
    ISP_PMGR_2 = 0x7f8, Register32
    ISP_PMGR_3 = 0x858, Register32

    ISP_ASC_RVBAR   = 0x1050000, Register64
    ISP_ASC_EDPRCR  = 0x1010310, Register32
    ISP_ASC_CONTROL = 0x1400044, Register32
    ISP_ASC_STATUS  = 0x1400048, Register32

    ISP_ASC_POWER_CYCLE_0 = 0x1400a00, Register32
    ISP_ASC_POWER_CYCLE_1 = 0x1400a04, Register32
    ISP_ASC_POWER_CYCLE_2 = 0x1400a08, Register32
    ISP_ASC_POWER_CYCLE_3 = 0x1400a0c, Register32
    ISP_ASC_POWER_CYCLE_4 = 0x1400a10, Register32
    ISP_ASC_POWER_CYCLE_5 = 0x1400a14, Register32

    ISP_REVISION = 0x1800000, ISP_REVISION

    ISP_POWER_UNK_0 = 0x20e0080, Register32
    ISP_POWER_UNK_1 = 0x20f0020, Register32
    ISP_POWER_UNK_2 = 0x20f8020, Register32

    ISP_GPIO_0 = 0x2104170, Register32
    ISP_GPIO_1 = 0x2104174, Register32
    ISP_GPIO_2 = 0x2104178, Register32
    ISP_GPIO_3 = 0x210417c, Register32
    ISP_GPIO_4 = 0x2104180, Register32
    ISP_GPIO_5 = 0x2104184, Register32
    ISP_GPIO_6 = 0x2104188, Register32
    ISP_GPIO_7 = 0x210418c, Register32

    ISP_GPIO_0_T8112 = 0x24c41b0, Register32
    ISP_GPIO_1_T8112 = 0x24c41b4, Register32
    ISP_GPIO_2_T8112 = 0x24c41b8, Register32
    ISP_GPIO_3_T8112 = 0x24c41bc, Register32
    ISP_GPIO_4_T8112 = 0x24c41c0, Register32
    ISP_GPIO_5_T8112 = 0x24c41c4, Register32
    ISP_GPIO_6_T8112 = 0x24c41c8, Register32
    ISP_GPIO_7_T8112 = 0x24c41cc, Register32

    ISP_SENSOR_CLOCK_0_EN = 0x2104190, Register32
    ISP_SENSOR_CLOCK_1_EN = 0x2104194, Register32
    ISP_SENSOR_CLOCK_2_EN = 0x2104198, Register32

    ISP_IRQ_INTERRUPT   = 0x2104000, Register32
    ISP_IRQ_ENABLE      = 0x2104004, Register32
    ISP_IRQ_DOORBELL    = 0x21043f0, Register32
    ISP_IRQ_ACK         = 0x21043fc, Register32
    ISP_IRQ_INTERRUPT_T8112 = 0x24c4000, Register32
    ISP_IRQ_DOORBELL_T8112 = 0x24c4430, Register32
    ISP_IRQ_ACK_T8112   = 0x24c443c, Register32

    ISP_IRQ_INTERRUPT_1 = 0x2104008, Register32
    ISP_IRQ_INTERRUPT_2 = 0x2104010, Register32
    ISP_IRQ_INTERRUPT_3 = 0x2104018, Register32
    ISP_IRQ_INTERRUPT_4 = 0x21043f8, Register32
    ISP_IRQ_INTERRUPT_4_T8112 = 0x24c4438, Register32

    ISP_DPE_UNK_0 = 0x2504000, Register32
    ISP_DPE_UNK_1 = 0x2508000, Register32

class ISPPSRegs(RegMap):  # This doesn't really make sense
    ISP_PS_00 = 0x4000, Register32
    ISP_PS_08 = 0x4008, Register32
    ISP_PS_10 = 0x4010, Register32
    ISP_PS_18 = 0x4018, Register32
    ISP_PS_20 = 0x4020, Register32
    ISP_PS_28 = 0x4028, Register32
    ISP_PS_30 = 0x4030, Register32
    ISP_PS_38 = 0x4038, Register32
    ISP_PS_40 = 0x4040, Register32
    ISP_PS_48 = 0x4048, Register32
    ISP_PS_50 = 0x4050, Register32
    ISP_PS_58 = 0x4058, Register32
    ISP_PS_60 = 0x4060, Register32



# SPDX-License-Identifier: MIT
"""
Things to note:
    The command buffer is encrypted after the poweron sequence, and I
    can't find the key in the SEP using sven's old SEP tracer.
"""
import struct
from construct import *

from m1n1.hv import TraceMode
from m1n1.proxyutils import RegMonitor
from m1n1.utils import *

from m1n1.trace import ADTDevTracer
from m1n1.trace.asc import ASCTracer, ASCRegs, EP, EPState, msg, msg_log, DIR
from m1n1.trace.dart import DARTTracer
from m1n1.trace.gpio import GPIOTracer
from m1n1.trace.spi import SPITracer

DARTTracer = DARTTracer._reloadcls()
ASCTracer = ASCTracer._reloadcls()
GPIOTracer = GPIOTracer._reloadcls()
SPITracer = SPITracer._reloadcls()

mesa_node = None
for node in hv.adt.walk_tree():
    try:
        if node.compatible[0] == "spi-1,spimc":
            for c in node:
                try:
                    if c.compatible[0] == "biosensor,mesa":
                        mesa_node = c
                        break
                except AttributeError:
                    continue
    except AttributeError:
        continue
    if mesa_node is not None:
        break

# trace interrupts
aic_phandle = getattr(hv.adt["/arm-io/aic"], "AAPL,phandle")
spi_node = mesa_node._parent

if getattr(spi_node, "interrupt-parent") == aic_phandle:
    for irq in getattr(spi_node, "interrupts"):
        hv.trace_irq("/arm-io/" + spi_node.name, irq, 1, hv.IRQTRACE_IRQ)

if getattr(mesa_node, "interrupt-parent") == aic_phandle:
    for irq in getattr(mesa_node, "interrupts"):
        hv.trace_irq("/arm-io/" + mesa_node.name, irq, 1, hv.IRQTRACE_IRQ)

mesa_pins = {
    0xc4: "mesa_pwr",

}


gpio_tracer = GPIOTracer(hv, "/arm-io/gpio0", mesa_pins, verbose=1)
gpio_tracer.start()

dart_sio_tracer = DARTTracer(hv, "/arm-io/dart-sio", verbose=1)
dart_sio_tracer.start()

iomon = RegMonitor(hv.u, ascii=True)

def readmem_iova(addr, size, readfn=None):
    try:
        return dart_sio_tracer.dart.ioread(0, addr, size)
    except Exception as e:
        print(e)
        return None

iomon.readmem = readmem_iova



class SIOMessage(Register64):
    EP    = 7, 0 # SPI2 DMA channels 0x18, 0x19
    TAG   = 13, 8 # counts up, message ID?
    TYPE  = 23, 16 # SIO message type
    PARAM = 31, 24
    DATA  = 63, 32

class SIOStart(SIOMessage):
    TYPE = 23, 16, Constant(2)

class SIOSetup(SIOMessage):
    TYPE = 23, 16, Constant(3)

class SIOConfig(SIOMessage): #???
    TYPE = 23, 16, Constant(5)

class SIOAck(SIOMessage):
    TYPE = 23, 16, Constant(0x65)

class SIOSetupIO(SIOMessage):
    TYPE = 23, 16, Constant(6)

class SIOCompleteIO(SIOMessage):
    TYPE = 23, 16, Constant(0x68)

class SIOEp(EP):
    BASE_MESSAGE = SIOMessage
    SHORT = "sioep"

    def __init__(self, tracer, epid):
        super().__init__(tracer, epid)
        self.state.iova = None
        self.state.iova_cfg = None
        self.state.iova_unk = None
        self.state.dumpfile = None

    @msg(2, DIR.TX, SIOStart)
    def Start(self, msg):
        self.log("Start SIO")

    @msg(3, DIR.TX, SIOSetup)
    def m_Setup(self, msg):
        if msg.EP == 0 and msg.PARAM == 0x1:
            self.state.iova = msg.DATA << 12

        elif msg.EP == 0 and msg.PARAM == 0x2:
            # size for PARAM == 0x1?
            iomon.add(self.state.iova, msg.DATA * 8,
                      name=f"SIO IOVA region at 0x{self.state.iova:08x}",
                      offset=self.state.iova)

        #elif msg.EP == 0 and msg.PARAM == 0xb:
            ## second iova block, maybe config
            #self.state.iova_cfg = msg.DATA << 12

        #elif msg.EP == 0 and msg.PARAM == 0xc:
            ## size for PARAM == 0xb?
            #iomon.add(self.state.iova_cfg, msg.DATA * 8,
                      #name=f"SIO IOVA CFG region at 0x{self.state.iova_cfg:08x}",
                      #offset=self.state.iova_cfg)

        if msg.EP == 0 and msg.PARAM == 0xd:
            # possible fingerprint sensor IOVA region
            self.state.iova_unk = msg.DATA << 12

        elif msg.EP == 0 and msg.PARAM == 0xe:
            iomon.add(self.state.iova_unk, msg.DATA * 8,
                      name=f"SIO IOVA UNK region at {self.state.iova_unk:08x}",
                      offset=self.state.iova_unk)

    @msg(5, DIR.TX, SIOConfig)
    def m_Config(self, msg):
        return

    @msg(0x65, DIR.RX, SIOAck)
    def m_Ack(self, msg):
        return

    @msg(6, DIR.TX, SIOSetupIO)
    def m_SetupIO(self, msg):
        if msg.EP == 0x18 or 0x19:
            iomon.poll()
        return

    @msg(0x68, DIR.RX, SIOCompleteIO)
    def m_CompleteIO(self, msg):
        if msg.EP == 0x18:
            if self.state.iova is None:
                return

            buf = struct.unpack("<I", self.tracer.ioread(self.state.iova + 0xa8, 4))[0]
            size = struct.unpack("<I", self.tracer.ioread(self.state.iova + 0xb0, 4))[0]
            self.log(f"SetupIO 0x18: buf {buf:#x}, size {size:#x}")
            # XXX: Do not try to log messages going to 0x2
            if buf == 0x2:
                self.log("Mesa command interrupted!")
                return
            self.log_mesa("EP 0x18", self.tracer.ioread(buf, size))
            return
        if msg.EP == 0x19:
            if self.state.iova is None:
                return

            buf = struct.unpack("<I", self.tracer.ioread(self.state.iova + 0x48, 4))[0]
            size = struct.unpack("<I", self.tracer.ioread(self.state.iova + 0x50, 4))[0]
            self.log(f"CompleteIO 0x19: buf {buf:#x}, size {size:#x}")
            # XXX: Do not try to log messages going to 0x2
            if buf == 0x2:
                self.log("Mesa command interrupted!")
                return
            if size >= 0x7200:
                with open("large_message.bin", "wb") as fd:
                    fd.write(self.tracer.ioread(buf, size))
                print("Fingerprint record message dumped.")
                return
            self.log_mesa("EP 0x19", self.tracer.ioread(buf, size))
            return

    def log_mesa(self, label, data):
        self.log(f"{label}: {len(data):d} byte message: ")
        chexdump(data)
        print("\n")


class SIOTracer(ASCTracer):
    ENDPOINTS = {
        0x20: SIOEp
    }


sio_tracer = SIOTracer(hv, "/arm-io/sio", verbose=1)
sio_tracer.start(dart=dart_sio_tracer.dart)

spi_tracer = SPITracer(hv, "/arm-io/spi2", verbose=2)
spi_tracer.start()



# SPDX-License-Identifier: MIT
import bisect, time

from .object import GPUObject, GPUAllocator
from .initdata import build_initdata
from .channels import *
from .event import GPUEventManager
from ..constructutils import Ver
from ..proxy import IODEV
from ..malloc import Heap
from ..hw.uat import UAT, MemoryAttr
from ..hw.agx import *
from ..fw.agx import AGXASC
from ..fw.agx.channels import ChannelInfoSet, ChannelInfo

class AGXChannels:
    pass

class AGXQueue:
    pass

class AGX:
    PAGE_SIZE = 0x4000
    MAX_EVENTS = 128

    def __init__(self, u):
        self.start_time = time.time()
        self.u = u
        self.p = u.proxy

        self.iface = u.iface
        self.show_stats = False

        self.asc_dev = u.adt["/arm-io/gfx-asc"]
        self.sgx_dev = u.adt["/arm-io/sgx"]
        if self.sgx_dev.compatible[0] in ("gpu,t6020", "gpu,t6021"):
            self.sgx = SGXRegsT602X(u, self.sgx_dev.get_reg(0)[0])
        else:
            self.sgx = SGXRegs(u, self.sgx_dev.get_reg(0)[0])

        self.log("Initializing allocations")

        self.aic_base = u.adt["/arm-io/aic"].get_reg(0)[0]

        self.all_objects = {}
        self.tracked_objects = {}

        # Memory areas
        self.fw_va_base = self.sgx_dev.rtkit_private_vm_region_base
        self.fw_va_size = self.sgx_dev.rtkit_private_vm_region_size
        self.kern_va_base = self.fw_va_base + self.fw_va_size

        # Set up UAT
        self.uat = UAT(self.u.iface, self.u)

        # Allocator for RTKit/ASC objects
        self.uat.allocator = Heap(self.kern_va_base + 0x80000000,
                                  self.kern_va_base + 0x81000000,
                                  self.PAGE_SIZE)

        self.asc = AGXASC(self.u, self.asc_dev.get_reg(0)[0], self, self.uat)
        self.asc.verbose = 0
        self.asc.mgmt.verbose = 0

        self.kobj = GPUAllocator(self, "kernel",
                                 self.kern_va_base, 0x20000000,
                                 AttrIndex=MemoryAttr.Shared, AP=1, guard_pages=16)
        self.cmdbuf = GPUAllocator(self, "cmdbuf",
                                   self.kern_va_base + 0x20000000, 0x20000000,
                                   AttrIndex=MemoryAttr.Shared, AP=1, UXN=1, PXN=1, guard_pages=16)
        self.kshared = GPUAllocator(self, "kshared",
                                    self.kern_va_base + 0x40000000, 0x20000000,
                                    AttrIndex=MemoryAttr.Shared, AP=1, guard_pages=16)
        self.kshared2 = GPUAllocator(self, "kshared2",
                                     self.kern_va_base + 0x60000000, 0x100000,
                                     AttrIndex=MemoryAttr.Shared, AP=0, PXN=1, guard_pages=16)
        self.kgpurw = GPUAllocator(self, "kernel GPU RW",
                                 self.kern_va_base + 0x70000000, 0x1000000,
                                 AttrIndex=MemoryAttr.Shared, AP=0, UXN=1, PXN=1)
        self.ktimestamp = GPUAllocator(self, "ktimestamp",
                                    self.kern_va_base + 0x71000000, 0x4000000,
                                    AttrIndex=MemoryAttr.Shared, AP=1, guard_pages=1)

        self.klow = GPUAllocator(self, "kernel_low",
                                 0x1500000000, 0x100000,
                                 AttrIndex=MemoryAttr.Shared, AP=0, UXN=1, PXN=1)
        self.klow.align_to_end = False

        self.io_allocator = Heap(self.kern_va_base + 0x68000000,
                                 self.kern_va_base + 0x70000000,
                                 block=self.PAGE_SIZE)

        self.mon = None
        self.event_mgr = GPUEventManager(self)

        self.p.iodev_set_usage(IODEV.FB, 0)

        self.initdata_hook = None
        self.halts_processed = 0

        # Early init, needed?
        self.poke_sgx()

    def poke_sgx(self):
        self.sgx_base = self.sgx_dev.get_reg(0)[0]
        self.p.read32(self.sgx_base + 0xd14000)
        self.p.write32(self.sgx_base + 0xd14000, 0x70001)

    def find_object(self, addr, ctx=0):
        all_objects = list(self.all_objects.items())
        all_objects.sort()

        idx = bisect.bisect_left(all_objects, ((ctx, addr + 1), "")) - 1
        if idx < 0 or idx >= len(all_objects):
            return None, None

        (ctx, base), obj = all_objects[idx]
        return base, obj

    def reg_object(self, obj, track=True):
        self.all_objects[(obj._ctx, obj._addr)] = obj
        if track:
            if self.mon is not None:
                obj.add_to_mon(self.mon)
            self.tracked_objects[(obj._ctx, obj._addr)] = obj

    def unreg_object(self, obj):
        del self.all_objects[(obj._ctx, obj._addr)]
        if obj._addr in self.tracked_objects:
            del self.tracked_objects[(obj._ctx, obj._addr)]

    def poll_objects(self):
        for obj in self.tracked_objects.values():
            diff = obj.poll()
            if diff is not None:
                self.log(diff)

    def alloc_channels(self, cls, name, channel_id, count=1, ring_size=0x100, rx=False):

        # All channels have 0x100 items
        item_count = ring_size
        item_size = cls.item_size()
        ring_size = item_count * item_size

        self.log(f"Allocating {count} channel(s) for {name} ({item_count} * {item_size:#x} bytes each)")

        state_obj = self.kshared.new_buf(0x30 * count, f"Channel.{name}.state", track=False)
        if rx:
            ring_buf = self.kshared.new_buf(ring_size * count, f"Channel.{name}.ring", track=False)
        else:
            ring_buf = self.kobj.new_buf(ring_size * count, f"Channel.{name}.ring", track=False)

        info = ChannelInfo()
        info.state_addr = state_obj._addr
        info.ringbuffer_addr = ring_buf._addr
        if name == "FWCtl":
            self.fwctl_chinfo = info
        else:
            setattr(self.ch_info, name, info)

        return [cls(self, name + ("" if count == 1 else f"[{i}]"), channel_id,
                    state_obj._paddr + 0x30 * i,
                    ring_buf._paddr + ring_size * i, item_count)
                for i in range(count)]

    def init_channels(self):
        self.log("Initializing channels...")
        self.ch_info = ChannelInfoSet()
        self.ch = AGXChannels()
        self.ch.queue = []

        # Command queue submission channels
        for index in range(4):
            queue = AGXQueue()
            self.ch.queue.append(queue)
            for typeid, chtype in enumerate(("TA", "3D", "CL")):
                name = f"{chtype}_{index}"
                chan = self.alloc_channels(GPUCmdQueueChannel, name,
                                           (index << 2) | typeid)[0]
                setattr(queue, "q_" + chtype, chan)

        # Device control channel
        self.ch.devctrl = self.alloc_channels(GPUDeviceControlChannel, "DevCtrl", 0x11)[0]

        # GPU -> CPU channels
        self.ch.event = self.alloc_channels(GPUEventChannel, "Event", None, rx=True)[0]
        self.ch.log = self.alloc_channels(GPULogChannel, "FWLog", None, 6, rx=True)
        self.ch.ktrace = self.alloc_channels(GPUKTraceChannel, "KTrace", None, ring_size=0x200, rx=True)[0]
        self.ch.stats = self.alloc_channels(GPUStatsChannel, "Stats", None, rx=True)[0]

        self.ch.fwctl = self.alloc_channels(GPUFWCtlChannel, "FWCtl", None, rx=False)[0]

        # For some reason, the FWLog channels have their rings in a different place...
        self.fwlog_ring = self.ch_info.FWLog.ringbuffer_addr
        self.ch_info.FWLog.ringbuffer_addr = self.kshared.buf(0x150000, "FWLog_Dummy")

    def poll_channels(self):
        for chan in self.ch.log:
            chan.poll()
        self.ch.ktrace.poll()
        if self.show_stats:
            self.ch.stats.poll()
        self.ch.event.poll()

    def kick_firmware(self):
        self.asc.db.doorbell(0x10)


    def get_irqs(self):
        hw_state = self.aic_base + 0x6800
        irqs = []
        for irq in self.sgx_dev.interrupts:
            v = int(bool((self.p.read32(hw_state + (irq // 32) * 4) & (1 << (irq % 32)))))
            irqs.append(v)
        return irqs

    def show_irqs(self):
        irqs = self.get_irqs()
        self.log(f' SGX IRQ state: {irqs}')

    def check_for_halt(self):
        irqs = self.get_irqs()
        if irqs[0]:
            if self.mon:
                self.mon.poll()
            self.poll_objects()
            self.log(f' SGX IRQ fired, checking for halt')
            self.wait_halt()
            self.log(r' (\________/) ')
            self.log(r'  |        |  ')
            self.log(r"'.| \  , / |.'")
            self.log(r'--| / (( \ |--')
            self.log(r".'|  _-_-  |'.")
            self.log(r'  |________|  ')
            self.log(r'')
            self.log(r' Halted nya~!!!!!')
            self.show_irqs()
            self.show_pending_stamps()
            self.check_fault()
            self.recover()

    def timeout(self, msg):
        self.log(msg)
        if msg.counter < self.halts_processed:
            self.log(f" Spurious, skipping")
            return

        if self.mon:
            self.mon.poll()
        self.poll_objects()
        self.log(msg)
        self.log(r' (\________/) ')
        self.log(r'  |        |  ')
        self.log(r"'.| \  , / |.'")
        self.log(r'--| / (( \ |--')
        self.log(r".'|  _-_-  |'.")
        self.log(r'  |________|  ')
        self.log(r'')
        self.log(r' Timeout nya~!!!!!')
        self.log(r'')
        self.log(f' Stamp index: {int(msg.stamp_index)}')
        self.show_pending_stamps()
        self.log(f' Fault info:')
        self.log(self.initdata.regionC.fault_info)

        self.show_irqs()
        self.check_fault()
        self.recover()

    def faulted(self, msg):
        if self.mon:
            self.mon.poll()
        self.poll_objects()
        self.log(msg)
        self.log(r' (\________/) ')
        self.log(r'  |        |  ')
        self.log(r"'.| \  , / |.'")
        self.log(r'--| / (( \ |--')
        self.log(r".'|  _-_-  |'.")
        self.log(r'  |________|  ')
        self.log(r'')
        self.log(r' Fault nya~!!!!!')
        self.log(r'')
        self.show_pending_stamps()
        self.log(f' Fault info:')
        self.log(self.initdata.regionC.fault_info)

        self.show_irqs()
        self.check_fault()
        self.recover()

    def show_pending_stamps(self):
        if Ver.check("V >= V13_5B4"):
            info_bits = 4
        else:
            info_bits = 3
        self.initdata.regionC.pull()
        self.log(f' Pending stamps:')
        for (off, i) in enumerate(self.initdata.regionC.pending_stamps):
            if i.info or i.wait_value:
                self.log(f"  - [{off}] #{i.info >> info_bits:3d}: {i.info & ((1 << info_bits) - 1)}/{i.wait_value:#x}")
            i.info = 0
            i.wait_value = 0
            tmp = i.regmap()
            tmp.info.val = 0
            tmp.wait_value.val = 0

        #self.initdata.regionC.push()

    def check_fault(self):
        fault_info = self.sgx.FAULT_INFO.reg
        if fault_info.value == 0xacce5515abad1dea:
            raise Exception("Got fault notification, but fault address is unreadable")

        self.log(f" Fault info: {fault_info}")

        if not fault_info.FAULTED:
            return

        fault_addr = getattr(self.sgx, "FAULT_ADDR", None)
        if fault_addr is not None:
            fault_addr = fault_addr.val << 6
        else:
            fault_addr = fault_info.ADDR << 6

        if fault_addr & 0x8000000000:
            fault_addr |= 0xffffff8000000000
        base, obj = self.find_object(fault_addr, ctx=fault_info.CONTEXT)
        info = ""
        if obj is not None:
            info = f" ({obj!s} + {fault_addr - base:#x})"
        self.log(f" GPU fault at {fault_addr:#x}{info}")
        self.log(f" Faulting unit: {agx_decode_unit(fault_info.UNIT)}")

    def wait_halt(self):
        status = self.fw_status
        while True:
            halted = bool(status.halted.val)
            self.log(f" Halted: {halted}")
            if halted:
                return

    def recover(self):
        status = self.fw_status
        self.log(f" Halt count: {status.halt_count.val}")
        halted = bool(status.halted.val)
        self.log(f" Halted: {halted}")
        if halted:
            self.log(f" Attempting recovery...")
            self.halts_processed = status.halt_count.val
            status.halted.val = 0
            status.resume.val = 1
        else:
            raise Exception("Cannot recover")
        self.show_irqs()

    def resume(self):
        self.log("Starting ASC")
        self.asc.start()

        self.log("Starting endpoints")
        self.asc.start_ep(0x20)
        self.asc.start_ep(0x21)

    def start(self):
        self.resume()

        self.init_channels()

        self.log("Building initdata")
        self.initdata = build_initdata(self)
        if self.initdata_hook:
            self.initdata_hook(self)

        self.fw_status = self.initdata.fw_status.regmap()
        self.uat.flush_dirty()

        self.log("Sending initdata")
        self.asc.fw.send_initdata(self.initdata._addr & 0xfff_ffffffff)
        self.asc.work()

        self.log("Sending DC_Init")
        self.ch.devctrl.send_init()
        self.asc.work()

        self.log("Sending DC_UpdateIdleTS")
        self.ch.devctrl.update_idle_ts()
        self.asc.work()

    def stop(self):
        self.asc.stop()

    def work(self):
        self.asc.work()

    def wait_for_events(self, timeout=1.0):
        now = time.time()
        deadline = now + timeout
        cnt = self.event_mgr.event_count
        while now < deadline and self.event_mgr.event_count == cnt:
            self.asc.work()
            now = time.time()
        if self.event_mgr.event_count == cnt:
            raise Exception("Timed out waiting for events")

    def log(self, msg):
        t = time.time() - self.start_time
        print(f"[AGX][{t:10.03f}] " + str(msg))
